#include "utility.h"
#include "openfile.h"
#include "directory.h"
#include "filehdr.h"

extern "C" void bzero(void *, unsigned int);
extern "C" void bcopy(const void *, void *, unsigned int);

// open directory.
Directory::Directory(OpenFile *dirFile)
{
    dirFilePtr = dirFile;
    ASSERT(dirFilePtr != NULL);
    ASSERT(dirFilePtr->Constructed() == TRUE);
    ASSERT(dirFilePtr->IsDir() == TRUE);
}

// Destroy directory instance
Directory::~Directory()
{
    delete dirFilePtr;
}

bool Directory::Initialize(int dotdot)
{
    ASSERT(dirFilePtr->Length() == 0);
    DirectoryEntry entry;
    bzero((char *)&entry, sizeof (entry));
    entry.inUse = TRUE;
    entry.sector = dirFilePtr->GetSector();
    entry.name[0] = '.';
    int ret = dirFilePtr->Write((char *)&entry, sizeof (entry));
    if (ret != sizeof (entry)) {
	printf("Directory #%d: cannot initialize '.'\n",
	       dirFilePtr->GetSector());
	dirFilePtr->Truncate();
	return (FALSE);
    }
    entry.inUse = TRUE;
    entry.sector = dotdot;
    entry.name[0] = entry.name[1] = '.';
    ret = dirFilePtr->Write((char *)&entry, sizeof (entry));
    if (ret != sizeof (entry)) {
	printf("Directory #%d: cannot initialize '..'\n", dotdot);
	dirFilePtr->Truncate();
	return (FALSE);
    }
    return (TRUE);
}

Directory * Directory::Clone()
{
    OpenFile *file = new OpenFile(dirFilePtr->GetSector());
    ASSERT(file != NULL);
    // file is already open, so clone must succeed
    ASSERT(file->Constructed() == TRUE);
    return (new Directory(file));
}

bool Directory::Deleted()
{
    return (dirFilePtr->IsDeleted());
}

// Find file name in directory, return the sector where its file header 
// is stored
int Directory::Find(char *name)
{
    if (dirFilePtr->IsDeleted())
	return (-1);
    dirFilePtr->Seek(0);
    int offset;
    for (offset = 0; offset < dirFilePtr->Length(); offset += DIR_ENTRY_SIZE) {
	DirectoryEntry entry;
	dirFilePtr->Read((char *)&entry, sizeof (entry));
	if (entry.inUse && strcmp(entry.name, name) == 0) {
	    return (entry.sector);
	}
    }
    return (-1);
}

// Add file name into directory
bool Directory::Add(char *name, int newSector)
{ 
    if (dirFilePtr->IsDeleted()) {
	// this only happens if we have this directory as current directory,
	// and the directory is being removed. fail operation.
	return (FALSE);
    }
    if (strlen(name) > FileNameMaxLen) {
	printf("Directory: %s: file name too long\n", name);
	return (FALSE);
    }
    int offset, freePos = -1;
    dirFilePtr->Seek(0);
    for (offset = 0; offset < dirFilePtr->Length(); offset += DIR_ENTRY_SIZE) {
	DirectoryEntry entry;
	dirFilePtr->Read((char *)&entry, sizeof (entry));
	if (entry.inUse && strcmp(entry.name, name) == 0) {
	    printf("Directory: %s: already exists (#%d)\n", name, entry.sector);
	    return (FALSE);
	} else if (!entry.inUse && freePos == -1)
	    freePos = offset;
    }
    if (freePos != -1) {
	ASSERT(freePos >= DIR_2_ENTRIES_SIZE);
	dirFilePtr->Seek(freePos);
    }
    DirectoryEntry entry;
    bzero((char *)&entry, sizeof (entry));
    entry.inUse = TRUE;
    entry.sector = newSector;
    strcpy(entry.name, name);
    int ret = dirFilePtr->Write((char *)&entry, sizeof (entry));
    if (ret == sizeof (entry))
	return (TRUE);
    else
	return (FALSE);
}

// Remove file name from directory
bool Directory::Remove(char *name)
{ 
    if (dirFilePtr->IsDeleted()) {
	// this only happens if we have this directory as current directory,
	// and the directory is being removed. fail operation.
	return (FALSE);
    }
    if (strcmp(name, "..") == 0) {
	printf("Remove: cannot remove '..'\n");
	return (FALSE);
    }
    if (strcmp(name, ".") == 0) {
	printf("Remove: cannot remove '.'\n");
	return (FALSE);
    }
    dirFilePtr->Seek(0);
    int offset;
    for (offset = 0; offset < dirFilePtr->Length(); offset += DIR_ENTRY_SIZE) {
	DirectoryEntry entry;
	dirFilePtr->Read((char *)&entry, sizeof (entry));
	if (entry.inUse && strcmp(entry.name, name) == 0) {
	    int sector = entry.sector;
	    // first try to remove the file.
	    if (CheckRemove(sector) == TRUE) {
		// then remove it from directory.
		entry.inUse = FALSE;
		entry.sector = -1;
		dirFilePtr->Seek(offset);
		dirFilePtr->Write((char *)&entry, sizeof (entry));
		return (TRUE);
	    } else {
		// FileHeader::Remove can only fail if the file is a directory,
		// and it is not empty.
		return (FALSE);
	    }
	}
    }
    printf("Remove %s: no such file or directory\n", name);
    return (FALSE);
}

bool Directory::CheckRemove(int sector)
{
    OpenFile *file = new OpenFile(sector, TRUE);
    if (!file->Constructed()) {
	delete file;
	return (FALSE);
    }
    if (!file->IsDir()) { /* regular file, can always be removed */
	file->FileWriteLock();
	file->MarkDeleted();
	file->FileWriteUnlock();
	delete file;
	return (TRUE);
    }
    file->DirWriteLock();
    bool activeEntries = FALSE;
    int offset = DIR_2_ENTRIES_SIZE;
    file->Seek(offset);
    while (offset < file->Length()) {
	DirectoryEntry entry;
	file->Read((char *)&entry, sizeof (entry));
	if (entry.inUse) {
	    activeEntries = TRUE;
	    break;
	}
	offset += DIR_ENTRY_SIZE;
    }
    bool canRemove;
    if (activeEntries) {
	canRemove = FALSE;
    } else {
	canRemove = TRUE;
	file->MarkDeleted();
    }
    file->DirWriteUnlock();
    delete file;
    return (canRemove);
}

int Directory::GetSector()
{
    return (dirFilePtr->GetSector());
}

// Debugging
void Directory::List(int indent, bool verbose)
{ 
    dirFilePtr->Seek(0);
    for (int i = 0; i < dirFilePtr->Length(); i += sizeof (DirectoryEntry)) {
	DirectoryEntry entry;
	dirFilePtr->Read((char *)&entry, sizeof (entry));
	for (int id = 0; id < indent; id++)
	    printf("     ");
	printf("%4d %10s", entry.sector, entry.name);
        if (entry.inUse) {
	    FileHeader *hdr = FileHeader::Get(entry.sector);
	    ASSERT(hdr != NULL);
	    bool isDir = hdr->IsDir();
	    printf("  %c  %6d bytes, %4d blocks\n",
		   (isDir ? 'd' : 'r'),
		   hdr->FileLength(),
		   hdr->FileBlocks());
            if (verbose) {
                hdr->Print();
            }
	    FileHeader::Put(hdr);
	    if (isDir && i >= sizeof (DirectoryEntry) * 2) {
		OpenFile *of = new OpenFile(entry.sector);
		ASSERT(of->Constructed() == TRUE);
		Directory *dir = new Directory(of);
		dir->List(indent+1, verbose);
		delete dir;
	    }
        } else
            printf("\t-\n");
        if (verbose)
            printf("\n");
    }
}
