#include "diskbitmap.h"
#include "diskcache.h"

extern "C" void bzero(void *, unsigned int);
extern "C" void bcopy(const void *, void *, unsigned int);

DiskBitMap::DiskBitMap(int nitems, int sector, bool initialize)  { 
  mutex = new Lock("BitMapLock");  
  ASSERT((nitems % BitsInWord) == 0);
  sectorNumber = sector;
  numBits = nitems;
  map = (unsigned int *)kernel->diskCache->Read(sectorNumber);
  dirty = initialize;
  
}

DiskBitMap::~DiskBitMap() {
  Update();
  kernel->diskCache->Release((char *)map, FALSE);
}

void DiskBitMap::Mark(int which)  { 
  ASSERT(which >= 0 && which < numBits);
  mutex->Acquire();
  map[which / BitsInWord] |= 1 << (which % BitsInWord);
  dirty = TRUE;
  mutex->Release();
}
    
void DiskBitMap::Clear(int which) {
    ASSERT(which >= 0 && which < numBits);
    mutex->Acquire();
    map[which / BitsInWord] &= ~(1 << (which % BitsInWord));
    dirty = TRUE;
    mutex->Release();
}

bool DiskBitMap::Test(int which) {
    ASSERT(which >= 0 && which < numBits);    
    if (map[which / BitsInWord] & (1 << (which % BitsInWord))) 
      return TRUE;
    return FALSE;
}

int DiskBitMap::Find()  {
  mutex->Acquire();
  for (int i = 0; i < numBits; i++)
    if ((map[i / BitsInWord] & (1 << (i % BitsInWord))) == 0) {
      map[i / BitsInWord] |= 1 << (i % BitsInWord);
      dirty = TRUE;
      mutex->Release();
      return (i);
    }
  mutex->Release();
  return (-1);
}

int DiskBitMap::NumClear()  {
    int count = 0;
    for (int i = 0; i < numBits; i++)
	if (!Test(i)) count++;
    return count;
}

void DiskBitMap::Print()  {
  printf("Bitmap set:\n"); 
  int rows = 0;
  for (int i = 0; i < numBits; i += BitsInWord) {
    printf("#%3d:%08x, ", i, map[i / BitsInWord]);
    if (++rows == 4) {
      printf("\n");
      rows = 0;
    }
  }
  printf("\n"); 
}

void DiskBitMap::Update() {
  mutex->Acquire();
  if (dirty) {
    kernel->diskCache->WriteAsync((char *)map);
    map = (unsigned int *)kernel->diskCache->Read(sectorNumber);
    dirty = FALSE;
  }
  mutex->Release();
}
