//
// diskcache.cc -- cache interface to disk.
//			
// features:
//  - synchronous reads
//  - asynchronous reads (for read ahead)
//  - asynchronous writes
//  - delayed writes
//
// all of this functionality is implemented by using a cache manager thread.
//

#include "synchdisk.h"
#include "diskcache.h"
#include "disk.h"
#include "utility.h"

extern "C" int write(int, char *, int);

Thread *gAsyncThread;

static void StartDaemon(int arg)
{
    printf("async I/O daemon started\n");
    kernel->diskCache->AsyncIoDaemon();
}

DiskCache::DiskCache()
{
  mutex = new Lock("CacheManager");
  wait = new Condition("CacheManager");
  waitIOreq = new Condition ("Async I/O Daemon");

    head.bufferNextPtr = (buffer_t *)&head;
    head.bufferPrevPtr = (buffer_t *)&head;
    for (int i = 0; i < NUM_BUFFERS; i++) {
	buffers[i].sectorNum = -1;
	buffers[i].dirty = FALSE;
	buffers[i].inuse = FALSE;
	buffers[i].ownerThreadPtr = NULL;
	InsTail(&buffers[i]);
    }
    hits = misses = waits = 0;
    printf("using %d disk cache buffers\n", NUM_BUFFERS);

    gAsyncThread = new Thread("Async I/O Daemon");
    gAsyncThread->Fork((VoidFunctionPtr)StartDaemon,(void *)0);
}

DiskCache::~DiskCache(void)
{
    write(1, "syncing disks ", 14);
    Sync();
    write(1, "\n", 1);
    for (int i = 0; i < NUM_BUFFERS; i++) {
	ASSERT(!buffers[i].dirty);
	ASSERT(!buffers[i].inuse);
	ASSERT(buffers[i].asycmd == ASY_NULL);
    }
    printf("disk cache: %d hits, %d misses, %d waits\n",
	   hits, misses, waits);
}

// Get a block, without reading from disk. This method is used when a block
// will be written (so we don't need its old contents).
char * DiskCache::Get(int sector)
{
    register buffer_t *bp = Lookup(sector);
    if (bp != NULL) {
	hits++;
    } else {
	misses++;
	bp = Evict();
	bp->sectorNum = sector;
    }
    ASSERT(bp->sectorNum == sector);
    ASSERT(bp->inuse == TRUE);
    ASSERT(bp->ownerThreadPtr == kernel->currentThread);
    return (bp->bufferData);
}

// Read a block.
char * DiskCache::Read(int sector)
{
    register buffer_t *bp = Lookup(sector);
    if (bp != NULL) {
	hits++;
    } else {
	misses++;
	bp = Evict();
	ASSERT(bp->dirty == FALSE);
	bp->sectorNum = sector;
	kernel->synchDisk->ReadSector(sector, bp->bufferData);
    }
    ASSERT(bp->sectorNum == sector);
    ASSERT(bp->inuse == TRUE);
    ASSERT(bp->ownerThreadPtr == kernel->currentThread);
    return (bp->bufferData);
}

// Start an asynchronous read of a block. Used for read ahead.
void DiskCache::ReadAsync(int sector)
{
    ASSERT(sector >= 0 && sector < NumSectors);
    register buffer_t *bp, *freebuf = NULL;
    mutex->Acquire();
    // first we must look for it in cache.
    bp = head.bufferPrevPtr;
    while (bp != (buffer_t *)&head) {
	if (bp->sectorNum == sector) {
	    // block is already in cache.
	    // nothing to do, just put it at the front.
	    Remove(bp);
	    InsHead(bp);
	    mutex->Release();
	    return;
	}
	if (bp->inuse == FALSE && bp->dirty == FALSE && freebuf == NULL)
	    freebuf = bp;
	bp = bp->bufferPrevPtr;
    }
    // block is not in cache.
    // if there is a free, non-dirty, unused buffer, start read ahead.
    if (freebuf != NULL) {
	ASSERT(freebuf->asycmd == ASY_NULL);
	ASSERT(freebuf->ownerThreadPtr == NULL);
	freebuf->inuse = TRUE;
	freebuf->asycmd = ASY_READ;
	freebuf->ownerThreadPtr = gAsyncThread;
	freebuf->sectorNum = sector;
	Remove(freebuf);
	InsHead(freebuf);
	ioQ.Prepend((buffer_t *)freebuf);
	waitIOreq->Broadcast(mutex);
    }
    mutex->Release();
}

// Release a block obtained with read or get.
// Return the block to the list of available blocks. Depending on the willReuse
// flag, it may be put at the front or at the tail of the LRU chain.
// Blocks that are expected to be needed again shortly
// (e.g. partially read data blocks) go to the front.
// Blocks that are unlikely to be needed again shortly go to the tail.
void DiskCache::Release(char *buf, bool willReuse)
{
    register buffer_t *bp = data2buf(buf);
    ASSERT(bp->inuse == TRUE);
    mutex->Acquire();
    bp->ownerThreadPtr = NULL;
    bp->inuse = FALSE;
    Remove(bp);
    if (willReuse) {
	InsHead(bp);
    } else {
	InsTail(bp);
    }
    wait->Broadcast(mutex);
    mutex->Release();
}

// Release and write a block. Just mark the block as dirty.
// Actual write will occur when the block is needed.
void DiskCache::WriteDelayed(char *buf, bool willReuse)
{
    register buffer_t *bp = data2buf(buf);
    bp->dirty = TRUE;
    Release(buf, willReuse);
}

// Write a block asynchronously.
void DiskCache::WriteAsync(char *buf)
{
    register buffer_t *bp = data2buf(buf);
    ASSERT(bp->inuse == TRUE);
    mutex->Acquire();
    bp->dirty = TRUE;
    bp->asycmd = ASY_WRITE;
    bp->ownerThreadPtr = gAsyncThread;
    Remove(bp);
    InsTail(bp);
    ioQ.Append((buffer_t *)bp);
    waitIOreq->Broadcast(mutex);
    mutex->Release();
}

void DiskCache::AsyncIoDaemon()
{
    mutex->Acquire();
    for (;;) {
      while(ioQ.IsEmpty()) waitIOreq->Wait(mutex);
      buffer_t *bp = (buffer_t *)ioQ.RemoveFront();      
      ASSERT(bp->sectorNum >= 0 && bp->sectorNum < NumSectors);
      ASSERT(bp->inuse == TRUE);
      ASSERT(bp->asycmd == ASY_READ || bp->asycmd == ASY_WRITE);
      ASSERT(bp->ownerThreadPtr == gAsyncThread);
      mutex->Release();
      if (bp->asycmd == ASY_READ) {
	kernel->synchDisk->ReadSector(bp->sectorNum, bp->bufferData);
      } else {
	ASSERT(bp->dirty == TRUE);
	kernel->synchDisk->WriteSector(bp->sectorNum, bp->bufferData);
      }
      mutex->Acquire();
      bp->dirty = FALSE;
      bp->inuse = FALSE;
      bp->asycmd = ASY_NULL;
      bp->ownerThreadPtr = NULL;
      wait->Broadcast(mutex);
    }
}

// Flush all cache blocks down to disk, for clean shutdown.
void DiskCache::Sync(void)
{
    mutex->Acquire();
    register buffer_t *bp;
    int n;
again:
    n = 0;
    bp = head.bufferNextPtr;
    while (bp != (buffer_t *)&head) {
	if (bp->asycmd != ASY_NULL) {
	    ASSERT(bp->inuse == TRUE);
	    wait->Wait(mutex);
	    goto again;
	}
	ASSERT(bp->asycmd == ASY_NULL);
	if (bp->dirty && !bp->inuse) {
	    ASSERT(bp->sectorNum >= 0 && bp->sectorNum < NumSectors);
	    ASSERT(bp->ownerThreadPtr == NULL);
	    kernel->synchDisk->WriteSector(bp->sectorNum, bp->bufferData);
	    bp->dirty = FALSE;
	}
	bp = bp->bufferNextPtr;
	n++;
    }
    wait->Broadcast(mutex);
    mutex->Release();
    ASSERT(n == NUM_BUFFERS);
}

buffer_t * DiskCache::Lookup(int sector)
{
    ASSERT(sector >= 0 && sector < NumSectors);
    register buffer_t *bp;
    mutex->Acquire();
    bp = head.bufferNextPtr;
    while (bp != (buffer_t *)&head) {
	if (bp->sectorNum == sector) {
	    if (bp->inuse) {
		ASSERT(bp->ownerThreadPtr != NULL);
		ASSERT(bp->ownerThreadPtr != kernel->currentThread);
		waits++;
		wait->Wait(mutex);
		// after waiting, everything might have changed.
		// play it again.
		bp = head.bufferNextPtr;
		continue;
	    }
	    ASSERT(bp->ownerThreadPtr == NULL);
	    bp->ownerThreadPtr = kernel->currentThread;
	    bp->inuse = TRUE;
	    mutex->Release();
	    return (bp);
	}
	bp = bp->bufferNextPtr;
    }
    mutex->Release();
    return (NULL);
}

buffer_t * DiskCache::Evict()
{
    register buffer_t *bp;
    mutex->Acquire();
    for (;;) {
	bp = head.bufferPrevPtr;
	while (bp != (buffer_t *)&head) {
	    if (!bp->inuse) {
		ASSERT(bp->ownerThreadPtr == NULL);
		ASSERT(bp->asycmd == ASY_NULL);
		bp->inuse = TRUE;
		bp->ownerThreadPtr = kernel->currentThread;
		mutex->Release();
		if (bp->dirty) {
		    kernel->synchDisk->WriteSector(bp->sectorNum, bp->bufferData);
		    bp->dirty = FALSE;
		}
		return (bp);
	    }
	    bp = bp->bufferPrevPtr;
	}
	// out of luck -- we must wait.
	waits++;
	wait->Wait(mutex);
    }
}
