//
// diskcache.h -- cache interface to disk.
//			
// Features:
//  - synchronous reads
//  - asynchronous reads (for read ahead)
//  - asynchronous writes
//  - delayed writes
//
// All of this functionality is implemented by using a cache manager thread.
//
// Cache blocks are acquired with Read()/Get(),
// and released with Release()/WriteDelayed()/WriteAsync().
// During this time, a block belongs to
// the caller, and must not be touched by the cache manager.
//
// On WriteDelayed(), the cache manager will mark the block as dirty: write
// will occur when the block is needed.
// When a block is Released or Written, is put in the front of the list if
// willReuse is FALSE, otherwise at the end.
//
// On WriteAsync(), the cache manager will immediately start an asynchronous
// write of the block. This call is made when the block will not be reused soon,
// so in the case the block will be put at the end of the free list.
//
// Free blocks are gotten from the front of the list, in this way the caller
// controls the way cache is managed. the 'willReuse' flag is effectively an
// advise to the cache manager on how likely the block will be reused in the
// future.
//
#ifndef _DISKCACHE_H_
#define _DISKCACHE_H_

#include "synch.h"
#include "list.h"

const int NUM_BUFFERS = 64;

// All buffers in buffer cache are organized in a doubly linked LRU
// list. The most recently used buffer will be at the head of the list,
// while the least recently used buffer will be at the tail of the list.
// A buffer is released with the indication of whether or not will be used
// again. If yes, it will be put at the front of the list, otherwise at the
// end of the list.
// When a cache miss occurs, new buffers will be taken from the tail of the
// list.
//
// We freely inspired (i.e., copied) the code from <sys/buf.h>, and we
// utilize the same scheme for implementing the double linked chain.
// We have a 'bufhead' structue, which is equal to the first fields of
// the buffer structure.
// 
// For some reason the compiler complains if we put this declaration inside
// the DiskCache class, where it belongs.
//

enum { ASY_NULL, ASY_READ, ASY_WRITE }; // async I/O command.

typedef struct buffer {
    struct buffer *bufferNextPtr;	// ptr to next buf in chain.
    struct buffer *bufferPrevPtr;	// ptr to previous buf in chain.
    int sectorNum;		// block number.
    bool dirty;		// buffer will be written to disk if dirty.
    bool inuse;		// true if buffer in use.
    int asycmd;
    Thread *ownerThreadPtr;		// debug.
    char bufferData[SectorSize];	// data itself - must be the last field
} buffer_t;

struct bufhead {
    buffer_t *bufferNextPtr;		// ptr to next buf in chain.
    buffer_t *bufferPrevPtr;		// ptr to previous buf in chain.
};

class DiskCache {
public:
    DiskCache(void);
    ~DiskCache(void);

    // Get a block, without reading from disk. This method is used when a block
    // will be written (so we don't need its old contents).
    char *Get(int sector);

    // Read a block.
    char *Read(int sector);

    // Start an asynchronous read of a block. Used for read ahead.
    void ReadAsync(int sector);

    // Release a block obtained with read or get.
    // willReuse is TRUE if the caller knows the block may be reused soon.
    void Release(char *buffer, bool willReuse);

    // Mark the block as dirty and release it.
    // willReuse is TRUE if the caller knows the block may be reused soon.
    void WriteDelayed(char *buffer, bool willReuse);
    // Start an asynchronous write of the block.
    void WriteAsync(char *buffer);

    // Flush all cache blocks down to disk.
    void Sync(void);

    // Entry point for async I/O daemon.
    void AsyncIoDaemon();

private:
    buffer_t buffers[NUM_BUFFERS];

    Lock *mutex;
    Condition *wait;		// processes wait on this condition for a
				// buffer to become available.
    Condition *waitIOreq;	// async I/O daemon waits on this condition
				// for requests.

    // get a buffer from the LRU chain.
    // If the buffer is already in the LRU chain, it will return it,
    // possibly after waiting for its release (if it is in use).
    // If the buffer is not in the LRU chain, NULL will be returned.
    buffer_t *Lookup(int sector);
    // evict a buffer from the LRU chain, possibly writing it to disk if
    // necessary.
    buffer_t *Evict();

    // pointers to the linked list.
    struct bufhead head;

    // pending I/O queue.
    List<buffer_t *> ioQ;

    buffer_t *data2buf(char *data) {
	buffer_t *bp = (buffer_t *)&data[SectorSize] - 1;
	ASSERT(bp >= buffers && bp < &buffers[NUM_BUFFERS]);
	ASSERT(&buffers[bp - buffers] == bp); /* make sure it's aligned */
	ASSERT(bp->sectorNum >= 0 && bp->sectorNum < NumSectors);
	return (bp);
    };
    /*
     * Insq/Remq for the buffer lists.
     */
    // Remove a block from LRU Chain.
    void Remove(register buffer_t *bp) {
	(bp)->bufferPrevPtr->bufferNextPtr = (bp)->bufferNextPtr;
	(bp)->bufferNextPtr->bufferPrevPtr = (bp)->bufferPrevPtr;
    }
    // Add a block to the front of the chain.
    void InsHead(register buffer_t *bp) {
	register buffer_t *dp = (buffer_t *)&head;
	(dp)->bufferNextPtr->bufferPrevPtr = (bp);
	(bp)->bufferNextPtr = (dp)->bufferNextPtr;
	(dp)->bufferNextPtr = (bp);
	(bp)->bufferPrevPtr = (dp);
    }
    // Add a block to the end of the chain.
    void InsTail(register buffer_t *bp) {
	register buffer_t *dp = (buffer_t *)&head;
	(dp)->bufferPrevPtr->bufferNextPtr = (bp);
	(bp)->bufferPrevPtr = (dp)->bufferPrevPtr;
	(dp)->bufferPrevPtr = (bp);
	(bp)->bufferNextPtr = (dp);
    };

    // statistics
    int	hits; // # of hits
    int misses; // # of misses
    int waits; // # of wait for buffers
};

extern class DiskCache *gDiskCache;

#endif
