#include "diskbitmap.h"
#include "filehdr.h"
#include "synchdisk.h"
#include "diskcache.h"
#include "directory.h"

extern DiskBitMap *gBitMap;

extern "C" void bzero(void *, unsigned int);
extern "C" void bcopy(const void *, void *, unsigned int);

FileHeader::FileHeader(int sector)
{
    hdrSector = sector;
    deleted = FALSE;
    busyCount = 0;
    ASSERT(hdrSector != -1);
    ASSERT(gBitMap->Test(hdrSector));
    FetchFrom();
    ASSERT(dirty == FALSE);
    lastLsector = -1;
    lastPsector = -1;
}

FileHeader::~FileHeader()
{
    ASSERT(header);
    ASSERT(busyCount == 0);
    if (deleted) {
	Deallocate();
    } else if (dirty) {
	kernel->diskCache->WriteDelayed((char *)header, TRUE);
    } else {
	kernel->diskCache->Release((char *)header, TRUE);
    }
}

// Initialize a fresh file header
void FileHeader::Initialize(bool isdir)
{ 
    ASSERT(header != 0);
    if (isdir)
        header->flags = FLAGS_DIR;
    else
	header->flags = FLAGS_REG;
    header->fileSize = 0;
    for (int i = 0; i < DIRECT_BLOCKS; i++)
	header->dataBlocks[i] = -1;
    header->indir1 = -1;
    header->indir2 = -1;
    header->blocks = 0;
    dirty = TRUE;
}

// truncate file size to 0
bool FileHeader::Truncate()
{
    ASSERT(header != 0);
    if (busyCount != 0) {
	return (FALSE);
    }
    if (header->fileSize == 0) {
	return (TRUE);
    }
    // deallocate L0 blocks.
    DeallocateBlocks(header->dataBlocks, DIRECT_BLOCKS);

    // deallocate L1 blocks.
    if (header->indir1 != -1) {
	ASSERT(gBitMap->Test(header->indir1));

	int *L1array = (int *)kernel->diskCache->Read(header->indir1);
	DeallocateBlocks(L1array, INDIR_BLOCKS);
	kernel->diskCache->Release((char *)L1array, FALSE);

	gBitMap->Clear(header->indir1);
	header->blocks--;

	header->indir1 = -1;
    }

    // deallocate L2 blocks.
    if (header->indir2 != -1) {
	ASSERT(gBitMap->Test(header->indir2));

	int *L2array = (int *)kernel->diskCache->Read(header->indir2);
	for (int i = 0; i < INDIR_BLOCKS; i++) {
	    if (L2array[i] == -1)
		continue;
	    ASSERT(gBitMap->Test(L2array[i]));

	    int *L2blocks = (int *)kernel->diskCache->Read(L2array[i]);
	    DeallocateBlocks(L2blocks, INDIR_BLOCKS);
	    kernel->diskCache->Release((char *)L2blocks, FALSE);

	    gBitMap->Clear(L2array[i]);
	    header->blocks--;
	}
	kernel->diskCache->Release((char *)L2array, FALSE);

	gBitMap->Clear(header->indir2);
	header->blocks--;

	header->indir2 = -1;
    }

    ASSERT(header->blocks == 0);
    header->fileSize = 0;
    lastLsector = -1;
    lastPsector = -1;
    dirty = TRUE;
    return (TRUE);
}

void FileHeader::DeallocateBlocks(int *array, int count)
{
    for (int i = 0; i < count; i++) {
	if (array[i] == -1)
	    continue;
	ASSERT(gBitMap->Test(array[i]));  // ought to be marked!
	gBitMap->Clear(array[i]);
	array[i] = -1;
	header->blocks--;
    }
}

void FileHeader::Deallocate()
{
    ASSERT(busyCount == 0);
    Truncate();
    ASSERT(header->blocks == 0);
    ASSERT(header->fileSize == 0);

    // we don't need to write it out.
    kernel->diskCache->Release((char *)header, FALSE);

    ASSERT(gBitMap->Test(hdrSector));
    gBitMap->Clear(hdrSector);
    dirty = FALSE;
    header = 0;
    hdrSector = -1;
}

int FileHeader::MapSectorWrite(int lOffset)
{
    ASSERT(header != 0);
    ASSERT(lOffset > 0);
    ASSERT(lOffset <= MaxFileSize);

    int currentLsector, lSector;
    currentLsector = lSector = (lOffset - 1) / SectorSize;

    if (lSector == lastLsector && lastPsector != -1) {
	if (lOffset > header->fileSize) {
	    header->fileSize = lOffset;
	    dirty = TRUE;
	}
	return (lastPsector);
    }

    // L0 indirection. --------------------------------------------------------

    if (lSector < DIRECT_BLOCKS) {
	// check whether data block exists. if not, we must allocate it.
	if (header->dataBlocks[lSector] == -1) {
	    header->dataBlocks[lSector] = AllocateDataBlock();
            if (header->dataBlocks[lSector] == -1)
	        return (-1);
	    dirty = TRUE;
	}

	if (lOffset > header->fileSize) {
	    header->fileSize = lOffset;
	    dirty = TRUE;
	}

        lastLsector = currentLsector;
	lastPsector = header->dataBlocks[lSector];
	return (header->dataBlocks[lSector]);
    }
    lSector -= DIRECT_BLOCKS;

    // L1 indirection. --------------------------------------------------------

    if (lSector < INDIR_BLOCKS) {
	// check whether L1 array exists. if not, we must create one.
	if (header->indir1 == -1) {
	    header->indir1 = AllocatePointerBlock();
	    if (header->indir1 == -1)
	        return (-1);
	    dirty = TRUE;
	}

        ASSERT(header->indir1 >= 0 && header->indir1 < NumSectors);
	ASSERT(gBitMap->Test(header->indir1));
	// check whether data block exists. if not, we must allocate it.
	int *L1array = (int *)kernel->diskCache->Read(header->indir1);
        int dataSector = L1array[lSector];
	if (dataSector == -1) {
	    dataSector = AllocateDataBlock();
	    if (dataSector == -1) {
		kernel->diskCache->Release((char *)L1array, TRUE);
		return (-1);
	    }
	    L1array[lSector] = dataSector;
	    kernel->diskCache->WriteDelayed((char *)L1array, TRUE);
	    dirty = TRUE;
	} else {
	    kernel->diskCache->Release((char *)L1array, TRUE);
	}

	if (lOffset > header->fileSize) {
	    header->fileSize = lOffset;
	    dirty = TRUE;
	}

        lastLsector = currentLsector;
	lastPsector = dataSector;

	return (dataSector);
    }
    lSector -= INDIR_BLOCKS;

    // L2 indirection. --------------------------------------------------------
    ASSERT(lSector < INDIR_BLOCKS * INDIR_BLOCKS);

    // check whether L2 array exists. if not, we must create one.
    if (header->indir2 == -1) {
	header->indir2 = AllocatePointerBlock();
	if (header->indir2 == -1)
	    return (-1);
	dirty = TRUE;
    }

    // read L2 array, and check for L2 block existence. we must create it
    // if it doesn't exist.
    ASSERT(header->indir2 >= 0 && header->indir2 < NumSectors);
    ASSERT(gBitMap->Test(header->indir2));
    int *L2array = (int *)kernel->diskCache->Read(header->indir2);
    int L2blockSector = L2array[lSector / INDIR_BLOCKS];
    if (L2blockSector == -1) {
	L2blockSector = AllocatePointerBlock();
	if (L2blockSector == -1) {
	    kernel->diskCache->Release((char *)L2array, TRUE);
	    return (-1);
	}
	L2array[lSector / INDIR_BLOCKS] = L2blockSector;
	kernel->diskCache->WriteDelayed((char *)L2array, TRUE);
	dirty = TRUE;
    } else
        kernel->diskCache->Release((char *)L2array, TRUE);

    // read L2 block, and check for data existence. we must create it
    // if it doesn't exist.
    ASSERT(L2blockSector >= 0 && L2blockSector < NumSectors);
    ASSERT(gBitMap->Test(L2blockSector));
    int *L2block = (int *)kernel->diskCache->Read(L2blockSector);
    int dataSector = L2block[lSector % INDIR_BLOCKS];
    if (dataSector == -1) {
	dataSector = AllocateDataBlock();
	if (dataSector == -1) {
	    kernel->diskCache->Release((char *)L2block, TRUE);
	    return (-1);
	}
	L2block[lSector % INDIR_BLOCKS] = dataSector;
	kernel->diskCache->WriteDelayed((char *)L2block, TRUE);
	dirty = TRUE;
    } else
        kernel->diskCache->Release((char *)L2block, TRUE);

    if (lOffset > header->fileSize) {
	header->fileSize = lOffset;
	dirty = TRUE;
    }

    lastLsector = currentLsector;
    lastPsector = dataSector;

    return (dataSector);
}

int FileHeader::AllocateSector(bool isData)
{
    int sector = gBitMap->Find();
    if (sector == -1) {
	printf("FileHeader #%d: out of disk space\n", hdrSector);
	return (-1);
    }
    char *buf = kernel->diskCache->Get(sector);
    if (isData) // data block - zero fill it.
        bzero(buf, SectorSize);
    else { // pointer block - fill it with -1's.
	for (int i = 0; i < SectorSize/sizeof(int); i++)
	    ((int *)buf)[i] = -1;
    }
    kernel->diskCache->WriteDelayed(buf, TRUE);
    header->blocks++;
    dirty = TRUE;
    return (sector);
}

// Fetch contents of header from disk 
void FileHeader::FetchFrom()
{
    header = (fileHeader *)kernel->diskCache->Read(hdrSector);
    dirty = FALSE;
}

// update contents of header to disk
void FileHeader::Update()
{
    ASSERT(hdrSector != -1);
    FileWriteLock();
    if (dirty) {
        kernel->diskCache->WriteAsync((char *)header);
        header = (fileHeader *)kernel->diskCache->Read(hdrSector);
	dirty = FALSE;
    } 
    FileWriteUnlock();
}

// Map logical sector to physical sector, possibly returning the read-ahead
// sector.
// if nextSector happens to be on a boundary between different arrays,
// we don't return it, because that would involve a considerable amount
// of extra I/O.
void FileHeader::MapSector(int lSector, int *sector, int *nextSector)
{
    ASSERT(header != 0);
    ASSERT(sector != 0);
    ASSERT(nextSector != 0);
    ASSERT(lSector * SectorSize < header->fileSize);
    *sector = -1;

    int currentLsector = lSector;

    if (lastLsector == lSector && lastPsector != -1) {
	*sector = lastPsector;
	return;
    }

    // L0 indirection.
    if (lSector < DIRECT_BLOCKS) {
	*sector = header->dataBlocks[lSector];
	if (lSector < DIRECT_BLOCKS - 1)
	    *nextSector = header->dataBlocks[lSector + 1];
	//printf("FileHeader #%d: L0 lsect %d --> %d (next %d)\n",
	//       hdrSector, lSector, *sector, *nextSector);
	lastLsector = currentLsector;
	lastPsector = *sector;
	return;
    }
    lSector -= DIRECT_BLOCKS;

    // L1 indirection.
    if (lSector < INDIR_BLOCKS) {
	// we must test for non-existence, since the file can contain holes.
	if (header->indir1 != -1) {
            ASSERT(header->indir1 >= 0 && header->indir1 < NumSectors);
            ASSERT(gBitMap->Test(header->indir1));
	    int *L1array = (int *)kernel->diskCache->Read(header->indir1);
	    *sector = L1array[lSector];
	    if (lSector < INDIR_BLOCKS - 1)
	        *nextSector = L1array[lSector + 1];
	    kernel->diskCache->Release((char *)L1array, TRUE);
	}
	//printf("FileHeader #%d: L1 lsect %d --> %d (next %d)\n",
	//       hdrSector, lSector, *sector, *nextSector);
	lastLsector = currentLsector;
	lastPsector = *sector;
	return;
    }
    lSector -= INDIR_BLOCKS;

    // L2 indirection.
    ASSERT(header->indir2 != -1);
    ASSERT(lSector < INDIR_BLOCKS * INDIR_BLOCKS);
    ASSERT(header->indir2 >= 0 && header->indir2 < NumSectors);
    ASSERT(gBitMap->Test(header->indir2));
    int *L2array = (int *)kernel->diskCache->Read(header->indir2);
    int L2sector = L2array[lSector / INDIR_BLOCKS];
    kernel->diskCache->Release((char *)L2array, TRUE);
    // we must test for non-existence, since the file can contain holes.
    if (L2sector != -1) {
	int *L2blocks = (int *)kernel->diskCache->Read(L2sector);
	*sector = L2blocks[lSector % INDIR_BLOCKS];
	if ((lSector % INDIR_BLOCKS) < INDIR_BLOCKS - 1)
	        *nextSector = L2blocks[lSector % INDIR_BLOCKS + 1];
        kernel->diskCache->Release((char *)L2blocks, TRUE);
    }
    //printf("FileHeader #%d: L2 lsect %d --> %d (next %d)\n",
    //       hdrSector, lSector, *sector, *nextSector);
    lastLsector = currentLsector;
    lastPsector = *sector;
}

// Return file length, in bytes
int FileHeader::FileLength()
{
    ASSERT(header != 0);
    return (header->fileSize);
}

int FileHeader::FileBlocks()
{
    ASSERT(header != 0);
    return (header->blocks);
}

bool FileHeader::IsDir()
{
    ASSERT(header != 0);
    return (header->flags == FLAGS_DIR ? TRUE : FALSE);
}

// for debugging
void FileHeader::Print()
{
    ASSERT(header != 0);
    CheckUsed(hdrSector);
    printf(" FileHeader #%d: File size: %d (%d blocks). File Type %s\n",
	   hdrSector, header->fileSize, header->blocks,
           IsDir() ? "directory" : "regular");
    printf(" L0 Blocks: ");
    int top;
    for (top = DIRECT_BLOCKS - 1; top > 0; top--) {
        if (header->dataBlocks[top] != -1)
            break;
    }
    int i;
    for (i = 0; i <= top ; i++) {
	printf("%d ", header->dataBlocks[i]);
        CheckUsed(header->dataBlocks[i]);
    }
    printf("\n");
    if (header->indir1 != -1) {
        printf(" L1 array (%d): ", header->indir1);
        int *L1array = (int *)kernel->diskCache->Read(header->indir1);
        for (top = INDIR_BLOCKS - 1; top > 0; top--) {
            if (L1array[top] != -1)
                break;
        }
        for (i = 0; i <= top; i++) {
            printf("%d ", L1array[i]);
            CheckUsed(L1array[i]);
	}
        kernel->diskCache->Release((char *)L1array, FALSE);
        printf("\n");
    } else
        printf(" L1 array: not allocated\n");
    if (header->indir2 != -1) {
        printf(" L2 array (%d): ", header->indir2);
        int *L2array = 0;
        L2array = (int *)kernel->diskCache->Read(header->indir2);
        for (top = INDIR_BLOCKS - 1; top > 0; top--) {
            if (L2array[top] != -1)
                break;
        }
        for (i = 0; i <= top; i++) {
            printf("%d ", L2array[i]);
            CheckUsed(L2array[i]);
	}
        printf("\n");
        for (i = 0; i < INDIR_BLOCKS; i++) {
            if (L2array[i] != -1) {
                printf("   L2 blocks #%d (%d): ", i, L2array[i]);
                int *L2blocks = 0;
                L2blocks = (int *)kernel->diskCache->Read(L2array[i]);
                for (top = INDIR_BLOCKS - 1; top > 0; top--) {
                    if (L2blocks[top] != -1)
                        break;
                }
                for (int j = 0; j <= top; j++) {
                    printf("%d ", L2blocks[j]);
                    CheckUsed(L2blocks[j]);
		}
                printf("\n");
                kernel->diskCache->Release((char *)L2blocks, FALSE);
            }
        }
        kernel->diskCache->Release((char *)L2array, FALSE);
    } else
        printf(" L2 array: not allocated\n");
}

void FileHeader::CheckUsed(int sector)
{
    if (sector == -1)
	return;
    if (!gBitMap->Test(sector)) {
	printf("FileHeader #%d: WARNING: sector #%d not set in bitmap\n",
	       hdrSector, sector);
    }
}

void FileHeader::SetBusy() { busyCount++; }

bool FileHeader::ResetBusy()
{
    busyCount--;
    ASSERT(busyCount >= 0);
}

bool FileHeader::IsBusy() { return (busyCount != 0 ? TRUE : FALSE); }


FileHeader::sharedHdr FileHeader::sharedHdrs[MAX_OPEN_FILES];
Lock FileHeader::sharedHdrMutex("sharedHdrMutex");
int FileHeader::sharedHdrFreeCount = MAX_OPEN_FILES;

FileHeader::sharedHdr * FileHeader::sharedLookup(int sector, int *freeSlot)
{
    int index = sector % MAX_OPEN_FILES;
    if (freeSlot)
	*freeSlot = -1;
    for (int i = 0; i < MAX_OPEN_FILES; i++) {
	struct sharedHdr *shdr = &sharedHdrs[(index + i) % MAX_OPEN_FILES];
	if (shdr->hdr != NULL && shdr->sector == sector)
	    return (shdr);
	if (freeSlot && shdr->hdr == NULL && *freeSlot == -1)
	    *freeSlot = (index + i) % MAX_OPEN_FILES;
    }
    return (NULL);
}

void FileHeader::Initialize(int sector, bool isDir)
{
    sharedHdrMutex.Acquire();
    // better not be there yet!
    ASSERT(FileHeader::sharedLookup(sector, NULL) == NULL);
    FileHeader *hdr = new FileHeader(sector);
    hdr->Initialize(isDir);
    delete hdr;
    sharedHdrMutex.Release();
}

FileHeader * FileHeader::Get(int sector, bool useReservedSlots)
{
    int freeSlot;

    sharedHdrMutex.Acquire();
    struct sharedHdr *shdr = FileHeader::sharedLookup(sector, &freeSlot);
    if (shdr) {
	// found an already existing in-core copy.
	ASSERT(shdr->hdr != NULL);
	ASSERT(shdr->sector == sector);
	ASSERT(shdr->sector == shdr->hdr->GetSector());
	shdr->refCount++;
	sharedHdrMutex.Release();
	return (shdr->hdr);
    }
    if (FileHeader::sharedHdrFreeCount < RESERVED_SLOTS && !useReservedSlots) {
	sharedHdrMutex.Release();
	return (NULL);
    }
    if (freeSlot != -1) {
	// found a free slot.
	FileHeader::sharedHdrFreeCount--;
	ASSERT(FileHeader::sharedHdrFreeCount >= 0);
	shdr = &sharedHdrs[freeSlot];
	ASSERT(shdr->hdr == NULL);
	shdr->sector = sector;
	shdr->refCount = 1;
	shdr->hdr = new FileHeader(sector);
	ASSERT(shdr->sector == shdr->hdr->GetSector());
	sharedHdrMutex.Release();
	return (shdr->hdr);
    }
    sharedHdrMutex.Release();
    printf("FileHeader: file table full\n");
    return (NULL);
}

void FileHeader::Put(FileHeader *hdr)
{
    sharedHdrMutex.Acquire();
    struct sharedHdr *shdr = FileHeader::sharedLookup(hdr->GetSector(), NULL);
    ASSERT(shdr != NULL);
    ASSERT(shdr->hdr == hdr);
    ASSERT(shdr->sector == shdr->hdr->GetSector());
    shdr->refCount--;
    if (shdr->refCount == 0) {
	delete shdr->hdr;
	shdr->hdr = NULL;
	FileHeader::sharedHdrFreeCount++;
	ASSERT(FileHeader::sharedHdrFreeCount > 0);
	ASSERT(FileHeader::sharedHdrFreeCount <= MAX_OPEN_FILES);
    }
    sharedHdrMutex.Release();
}

void FileHeader::Sync()
{
    sharedHdrMutex.Acquire();
    for (int i = 0; i < MAX_OPEN_FILES; i++) {
	if (sharedHdrs[i].hdr != NULL)
	    sharedHdrs[i].hdr->Update();
    }
    sharedHdrMutex.Release();
}
