#ifndef FILEHDR_H
#define FILEHDR_H

#include "utility.h"
#include "synch.h"
#include "disk.h"
#include "diskcache.h"

// the max # of opened files must be less than the # of buffers, since
// we keep one buffer for each opened file, for an in-core copy of its
// buffer header.

const int MAX_OPEN_FILES = NUM_BUFFERS / 2;
const int RESERVED_SLOTS = MAX_OPEN_FILES / 4;

// We have three different levels of indirection:
// no indirection (direct blocks, level 0 - L0)
// single indirection (level 1 - L1)
// double indirection (level 2 - L2)

const int DIRECT_BLOCKS	= SectorSize / sizeof (int) - 5;
const int INDIR_BLOCKS = SectorSize / sizeof (int);
const int MAX_FILE_SIZE = DIRECT_BLOCKS +		      
			  INDIR_BLOCKS +		      
			  (INDIR_BLOCKS * INDIR_BLOCKS);  
const int MaxFileSize = MAX_FILE_SIZE * SectorSize; 

typedef struct {
    int flags;			   // 0=reg, 1=directory
    int fileSize;		   // size of file, in bytes.
    int blocks;			   // size of file, in blocks, including
				   // indirect pointers.
    int dataBlocks[DIRECT_BLOCKS]; // direct data blocks.
    int indir1;			   // pointer to level 1 indirect blocks.
    int indir2;			   // pointer to array of pointers to
				   // level 2 indirect blocks.
} fileHeader;
#define FLAGS_REG 0
#define FLAGS_DIR 1

class FileHeader {
    // caller must always use Get/Put.
    // see comment near the end of this class.
    FileHeader(int sector);
    ~FileHeader();

public:
    static void Initialize(int sector, bool isDir);
    static FileHeader * Get(int sector, bool useReservedSlots = FALSE);
    static void Put(FileHeader *hdr);
    static void Sync();

    void Update();

    bool Truncate();     // truncate file size to 0.

    void SetBusy();      // mark file busy.
    bool ResetBusy();
    bool IsBusy();

    // Map logical sector num to sector num, possibly
    // returning next sector num (for read-ahead).

    void MapSector(int lSector, int *sector, int *nextSector);

    // This is like MapSector, except that the sector is created if it
    // doesn't exist. We create a new sector for file, at logical byte
    // offset lByteOffset and return the physical sector number.
    // If the sector already exists, no allocation is needed, and its value
    // is returned. We need the byte offset, instead of a sector offset
    // in case the file needs to be extended.

    int MapSectorWrite(int lByteOffset);

    int FileLength();			// Return length of file in bytes.
    int FileBlocks();			// Return blocks of file.
    bool IsDir();			// TRUE if file is a directory.

    void Print(); // DEBUG

    int GetSector() { return (hdrSector); };

    // FileHeader has three types of locks:
    // - FileLock ReadWriteLock:
    //   used by OpenFile/Directory to synchronize multiple threads access.
    // - DirLock ReadWriteLock:
    //   used by Directory to synchronize multiple directory accesses.
    //	 Since a directory cannot be written with regular file I/O,
    //   atomicity is guaranteed.

    void FileReadLock() { FileLock.ReadLock(); };
    void FileReadUnlock() { FileLock.ReadUnlock(); };
    void FileWriteLock() { FileLock.WriteLock(); };
    void FileWriteUnlock() { FileLock.WriteUnlock(); };
    void DirReadLock() { DirLock.ReadLock(); };
    void DirReadUnlock() { DirLock.ReadUnlock(); };
    void DirWriteLock() { DirLock.WriteLock(); };
    void DirWriteUnlock() { DirLock.WriteUnlock(); };

    void MarkDeleted() { deleted = TRUE; };	// mark file as deleted.
    bool IsDeleted() { return (deleted); };

 private:
    void FetchFrom();			// fetch from disk.
    void Initialize(bool isDir);	// first time initialization.

    void Deallocate();  		// de-allocate file header.
    void DeallocateBlocks(int *array, int count);

    // This routine is used by MapSectorWrite to allocates either data
    // or an array of block pointers.  If isData is TRUE, the new
    // block is zero-filled, otherwise it is filled with -1.

    int AllocateSector(bool isData);
    int AllocateDataBlock() { return (AllocateSector(TRUE)); };
    int AllocatePointerBlock() { return (AllocateSector(FALSE)); };

    void CheckUsed(int sector); // DEBUG

    int hdrSector;
    bool dirty;
    fileHeader *header;
    bool deleted;
    int busyCount;

    int lastLsector, lastPsector;

    ReadWriteLock FileLock;
    ReadWriteLock DirLock;

    // This struct contains the in-core copies of all open file
    // headers.  When a file is kept open more than once, only one
    // copy is kept, via a reference count. When a file is deleted, it
    // is marked as deleted. When the reference count drops to zero
    // (i.e., every process closes it), the file is physically
    // removed. The caller calls Get()/Put() to get and release a file
    // header.

    static struct sharedHdr {
        FileHeader *hdr;
        int sector;
        int refCount;
    } sharedHdrs[MAX_OPEN_FILES];
    static int sharedHdrFreeCount;
    static struct sharedHdr *sharedLookup(int sector, int *freeSlot);
    static Lock sharedHdrMutex;
};

#endif
