// filesys.cc -- Routines to manage naming in the file system.
// ----------------------------------------------------------------------------
// NOTE: All code is changed, so we didn't use #ifdef in this file.
// Lot of original comments blown away as they are inaccurate now
// ----------------------------------------------------------------------------

#include <string.h>
#include "disk.h"
#include "diskbitmap.h"
#include "directory.h"
#include "filehdr.h"
#include "openfile.h"
#include "filesys.h"
#include "diskcache.h"

// Constants for the directory and bitmap files.
const int FREE_MAP_SECTORS = (NumSectors / BitsInByte) / SectorSize;
// Put these in a known place, so we can find them on boot up.
const FREE_MAP_SECTOR = 0;
const DIRECTORY_SECTOR = FREE_MAP_SECTOR + FREE_MAP_SECTORS;

DiskBitMap *gBitMap = 0;

extern "C" void bcopy(const void *, void *, unsigned int);

// Initialize the file system.
//   Open the files representing the directory and the bitmap of free disk 
//   blocks.  If we are "formatting" the disk (i.e., if the disk 
//   has garbage on it), we have to set up the disk first.
FileSystem::FileSystem( bool format )
{ 
    mutex = new Lock("FileSysMutex");

    if (format) {
	printf("Formatting file system\n");

	// initialize bitmap, accounting for itself.
	printf("initializing bitmap...\n");
        gBitMap = new DiskBitMap(NumSectors, FREE_MAP_SECTOR, TRUE);
	for (int i = 0; i < FREE_MAP_SECTORS; i++)
	    gBitMap->Mark(FREE_MAP_SECTOR + i);

	// Create the root directory

	printf("Creating root directory... ");
	//ASSERT(!gBitMap->Test(DIRECTORY_SECTOR));
	gBitMap->Mark(DIRECTORY_SECTOR);
	FileHeader::Initialize(DIRECTORY_SECTOR, TRUE);
	OpenFile *file = new OpenFile(DIRECTORY_SECTOR);
	ASSERT(file->Constructed() == TRUE);
	m_RootDirectory = new Directory(file);
	bool result = m_RootDirectory->Initialize(DIRECTORY_SECTOR);
	ASSERT(result == TRUE);
	printf("Done.\n");

	// create /swap directory.
	printf("Creating /swapdir directory... ");
	int swapSector = gBitMap->Find();
	ASSERT(swapSector != -1);
	FileHeader::Initialize(swapSector, TRUE);
	Directory *swapDir = new Directory(new OpenFile(swapSector));
	swapDir->Initialize(DIRECTORY_SECTOR);
	delete swapDir;
	m_RootDirectory->Add("swapdir", swapSector);
	printf("Done.\n");

	// create /bin directory.
	printf("Creating /bin directory... ");
	int binSector = gBitMap->Find();
	ASSERT(binSector != -1);
	FileHeader::Initialize(binSector, TRUE);
	Directory *binDir = new Directory(new OpenFile(binSector));
	binDir->Initialize(DIRECTORY_SECTOR);
	delete binDir;
	m_RootDirectory->Add("bin", binSector);
	printf("Done.\n");

	// create /usr/bin directory.
	printf("Creating /usr directory... ");
	int usrSector = gBitMap->Find();
	ASSERT(usrSector != -1);
	FileHeader::Initialize(usrSector, TRUE);
	Directory *usrDir = new Directory(new OpenFile(usrSector));
	usrDir->Initialize(DIRECTORY_SECTOR);
	m_RootDirectory->Add("usr", usrSector);
	printf("Done.\n");

	printf("Creating /usr/bin directory... ");
	int usrbinSector = gBitMap->Find();
	ASSERT(usrbinSector != -1);
	FileHeader::Initialize(usrbinSector, TRUE);
	Directory *usrbinDir = new Directory(new OpenFile(usrbinSector));
	usrbinDir->Initialize(usrbinSector);
	usrDir->Add("bin", usrbinSector);
	printf("Done.\n");

	delete usrDir;
	delete usrbinDir;

	// create /var/tmp directory.
	printf("Creating /var directory... ");
	int varSector = gBitMap->Find();
	ASSERT(varSector != -1);
	FileHeader::Initialize(varSector, TRUE);
	Directory *varDir = new Directory(new OpenFile(varSector));
	varDir->Initialize(DIRECTORY_SECTOR);
	m_RootDirectory->Add("var", varSector);
	printf("Done.\n");

	printf("Creating /var/tmp directory... ");
	int vartmpSector = gBitMap->Find();
	ASSERT(vartmpSector != -1);
	FileHeader::Initialize(vartmpSector, TRUE);
	Directory *tmpDir = new Directory(new OpenFile(vartmpSector));
	tmpDir->Initialize(varSector);
	varDir->Add("tmp", vartmpSector);
	printf("Done.\n");

	delete tmpDir;
	delete varDir;

    } else {
	gBitMap = new DiskBitMap(NumSectors, FREE_MAP_SECTOR, FALSE);
	for (int i = 0; i < FREE_MAP_SECTORS; i++)
	    ASSERT(gBitMap->Test(FREE_MAP_SECTOR + i));
	ASSERT(gBitMap->Test(DIRECTORY_SECTOR));
	OpenFile *file = new OpenFile(DIRECTORY_SECTOR);
	ASSERT(file->Constructed() == TRUE);
	m_RootDirectory = new Directory(file);
        int dot = m_RootDirectory->Find(".");
        ASSERT(dot == DIRECTORY_SECTOR);
        int dotdot = m_RootDirectory->Find("..");
        ASSERT(dotdot == DIRECTORY_SECTOR);
    }
}

FileSystem::~FileSystem()
{
    Sync();
    delete m_RootDirectory;
    delete gBitMap;
    delete mutex;
}

// Create a file.
OpenFile * FileSystem::Create(Directory *currentDir, char *pathname)
{
    OpenFile *of;
    bool result = CommonOpen(currentDir, pathname, OP_CREATE, &of);
    if (result == TRUE) {
	ASSERT(of->IsDir() == FALSE);
    }
    return (of);
}

// Create a directory.
bool FileSystem::MkDir(Directory *currentDir, char *pathname)
{
    OpenFile *of;
    bool result = CommonOpen(currentDir, pathname, OP_MKDIR, &of);
    ASSERT(of == NULL);
    return (result);
}

// Open a file.
OpenFile * FileSystem::Open(Directory *currentDir, char *pathname)
{
    OpenFile *of;
    bool result = CommonOpen(currentDir, pathname, OP_OPEN, &of);
    if (result == TRUE) {
	ASSERT(of != NULL);
    } else {
	ASSERT(of == NULL);
    }
    return (of);
}

// Change current directory.
Directory * FileSystem::ChDir(Directory *currentDir, char *pathname)
{

  if(!currentDir) {
    printf("FS: ChDir: NULL current directory.\n");
    return NULL;
  }

  if(!pathname) {
    printf("FS: ChDir: NULL pathname.\n");
    return NULL;
  }

    OpenFile *of;
    Directory *dir = NULL;
    bool result = CommonOpen(currentDir, pathname, OP_CHDIR, &of);
    if (result == TRUE) {
	if (of->IsDir()) {
	    dir = new Directory(of);
	} else {
	    printf("ChDir %s: not a directory\n", pathname);
	    delete of;
	}
    }
    return (dir);
}

bool FileSystem::CommonOpen(Directory *currentDir,
			    char *pathname,
			    op_t operation,
			    OpenFile **of)
{
    // currentDir can be NULL, which means root directory.
    // currentDir is NULL when
    // calls to file system are made by internal nachos threads,
    // and by address space swapfile create/remove calls.
    if (currentDir == NULL)
	currentDir = m_RootDirectory;
    *of = NULL;
    char name[FileNameMaxLen + 1];
    Directory *lastDir; // last directory in path before 'name'
    bool result = FollowPath(currentDir, pathname, &lastDir, name);
    if (result == FALSE) {
	ASSERT(lastDir == NULL);
	return (FALSE);
    }
    if (name[0] == 0) { /* root */
	delete lastDir;
	switch (operation) {
	    case OP_OPEN:
	    case OP_CHDIR:
		*of = new OpenFile(DIRECTORY_SECTOR);
		// root directory is always in file header table,
		// so there can be no overflow.
		//ASSERT(of->Constructed() == TRUE);
		return (TRUE);
	    case OP_CREATE:
		printf("MkDir %s: already exists\n", pathname);
		return (FALSE);
	    case OP_MKDIR:
		printf("Create %s: is a directory\n", pathname);
		return (FALSE);
	}
    }

    // we got a directory and the last component. now we must scan for it.
    ASSERT(lastDir != NULL);
    if (operation == OP_OPEN || operation == OP_CHDIR)
	lastDir->DirReadLock();
    else
	lastDir->DirWriteLock();

    result = FALSE;
    switch (operation) {
	case OP_OPEN:
	case OP_CHDIR: {
	    int sector = lastDir->Find(name);
	    if (sector == -1) {
		if (operation == OP_CHDIR)
		    printf("ChDir %s: directory does not exists\n", pathname);
		else
		    printf("Open %s: no such file or directory\n", pathname);
		break;
	    }
	    OpenFile *file = new OpenFile(sector);
	    if (file->Constructed() == FALSE) { // file header table overflow
		delete file;
		break;
	    }
	    *of = file;
	    result = TRUE;
	    break;
	}
	case OP_CREATE: {
	    if (lastDir->Deleted()) {
		// someone else is just removing this directory.
		printf("Create %s: no such file or directory\n", pathname);
		break;
	    }
	    int sector = lastDir->Find(name);
	    if (sector == -1) {
		sector = gBitMap->Find();
		if (sector == -1) {
		    printf("Create %s: out of disk space\n", pathname);
		    break;
		}
		FileHeader::Initialize(sector, FALSE);
		if (!lastDir->Add(name, sector)) {
		    // ran out of disk space.
		    gBitMap->Clear(sector);
		    break;
		}
	    }
	    OpenFile *file = new OpenFile(sector);
	    if (file->Constructed() == FALSE) {
		// file header table overflow -- undo the job
		lastDir->Remove(name);
		delete file;
	    } else if (file->IsDir()) {
		printf("Create %s: file is a directory\n", pathname);
		delete file;
	    } else if (file->IsBusy()) {
		printf("Create %s: paging file busy\n", pathname);
		delete file;
	    } else {
		file->Truncate();
		*of = file;
		result = TRUE;
	    }
	    break;
	}
	case OP_MKDIR: {
	    if (lastDir->Deleted()) {
		// someone else is just now removing this directory.
		printf("Create %s: no such file or directory\n", pathname);
		break;
	    }
	    int sector = lastDir->Find(name);
	    if (sector >= 0) {
		printf("Mkdir %s: already exists\n", pathname);
		break;
	    }
	    sector = gBitMap->Find();
	    if (sector == -1) {
		printf("Mkdir %s: out of disk space\n", pathname);
		break;
	    }
	    FileHeader::Initialize(sector, TRUE);
	    if (!lastDir->Add(name, sector)) {
		// ran out of disk space.
		gBitMap->Clear(sector);
		break;
	    }
	    OpenFile *file = new OpenFile(sector);
	    if (file->Constructed() == FALSE) {
		// file header table overflow -- undo the job
		delete file;
		lastDir->Remove(name);
	    } else {
		// the parent directory is locked, so we don't need to lock
		// the new directory.
		Directory *newDir = new Directory(file);
		if (newDir->Initialize(lastDir->GetSector()))
		    result = TRUE;
		else
		    // initialization failed because we run out of disk space,
		    // so we must undo our job.
		    lastDir->Remove(name);
		delete newDir;
		// OP_MKDIR does not return an open file.
	    }
	    break;
	}
    }

    if (operation == OP_OPEN || operation == OP_CHDIR)
	lastDir->DirReadUnlock();
    else
	lastDir->DirWriteUnlock();
    delete lastDir;
    return (result);
}

// Delete a file.  First, remove it from the directory, then delete its space.
bool FileSystem::Remove(Directory *currentDir, char *pathname)
{ 
    char name[FileNameMaxLen + 1];
    Directory *lastDir;
    // currentDir can be NULL, which means root directory.
    // currentDir is NULL when
    // calls to file system are made by internal nachos threads,
    // and by address space swapfile create/remove calls.
    if (currentDir == NULL)
	currentDir = m_RootDirectory;
    bool result = FollowPath(currentDir, pathname, &lastDir, name);
    if (result == FALSE) {
	ASSERT(lastDir == NULL);
	return (FALSE);
    }
    if (name[0] == 0) { /* remove root */
	delete lastDir;
	printf("Remove %s: cannot remove root directory\n", pathname);
	return (FALSE);
    }
    ASSERT(lastDir != NULL);
    lastDir->DirWriteLock();
    result = lastDir->Remove(name);
    if (result == FALSE) {
	// if the name exists in the directory, remove can only fail because
	// the name being removed is a non-empty directory.
	if (lastDir->Find(name) != -1)
	    printf("Remove %s: directory not empty\n", pathname);
	else
	    printf("Remove %s: no such file or directory\n", pathname);
    }
    lastDir->DirWriteUnlock();
    delete lastDir;
    return (result);
}


// clone root directory, for 1st process.
Directory * FileSystem::CloneRootDir()
{
    return (m_RootDirectory->Clone());
}

// clone a directory, for Exec().
Directory * FileSystem::CloneDir(Directory *dir)
{
    return (dir->Clone());
}

// this method follows a pathname 'path', starting from directory 'start'.
// If successful, this method returns TRUE, and 'dir' will be initialized
// with the last directory, while 'lastComponent' will contain the last part
// of the pathname.
bool FileSystem::FollowPath(Directory *start, char *path,
			    Directory **dir, char *lastComponent)
{
    *dir = NULL;
    *lastComponent = 0;
    ASSERT(start != NULL);
    ASSERT(path != NULL);
    ASSERT(lastComponent != NULL);
    //printf("FollowPath: path=|%s|\n",path);

    if (*path == 0) {
      printf("FollowPath: null pathname\n");
      return (FALSE);
    }

    if (*path == '/') {
      //printf("FollowPath: absolute pathname\n");
      start = m_RootDirectory;
      while (*path && *path == '/')
	path++;
    }

    //printf("FollowPath: dir#%d, path='%s'\n", start->GetSector(), path);
    if (*path == 0) { // handle "/" pathname.
      //printf("FollowPath: '/' pathname\n");
      *dir = m_RootDirectory->Clone();
      return (TRUE);
    }
    // to keep scan algorithm simple, we disallow pathnames ending with '/' --
    // except of course the pathname '/'.
    if (path[strlen(path) - 1] == '/') {
	printf("FollowPath: illegal pathname %s\n", path);
	return (FALSE);
    }

    //printf("FollowPath: Scaning path.\n");

    char *currComp = path;
    char *nextComp;
    Directory *currDir = start->Clone();
    for (;;) {
	ASSERT(*currComp != '/');
	char name[FileNameMaxLen + 1];
	nextComp = strchr(currComp, '/');
	int currCompLen = (nextComp != NULL) ? (nextComp - currComp) :
							    strlen(currComp);
	if (currCompLen > FileNameMaxLen) {
	    printf("FollowPath: path component of %s too long\n", path);
	    delete currDir;
	    return (FALSE);
	}
	bcopy(currComp, name, currCompLen);
	name[currCompLen] = 0;
	if (nextComp == NULL) {
	    strcpy(lastComponent, name);
	    *dir = currDir;
	    return (TRUE);
	}

	ASSERT(*nextComp == '/');
	while (*nextComp == '/')
	    nextComp++;
	currDir->DirReadLock();
	int sector = currDir->Find(name);
	if (sector == currDir->GetSector()) {
	    currComp = nextComp;
	    currDir->DirReadUnlock();
	    continue;
	}
	if (sector == -1) {
	    printf("FollowPath: path component %s: no such file or directory\n",
		   name);
	    currDir->DirReadUnlock();
	    delete currDir;
	    return (FALSE);
	}
	OpenFile *file = new OpenFile(sector);
	currDir->DirReadUnlock();
	if (file->Constructed() == FALSE) { // file header table overflow.
	    delete file;
	    delete currDir;
	    return (FALSE);
	}
	if (file->IsDir() == FALSE) {
	    printf("FollowPath: path component %s: not a directory\n",
		   name);
	    delete currDir;
	    return (FALSE);
	}
	Directory *nextDir = new Directory(file);
	delete currDir;
	currDir = nextDir;
	currComp = nextComp;
    }
}

extern "C" int write(int, char *, int);

void FileSystem::Sync()
{
    write(1, "syncing files,", 14);
    FileHeader::Sync();
    write(1, "bitmap,", 7);
    gBitMap->Update();
    write(1, "buffers", 7);
    kernel->diskCache->Sync();
    write(1, "\n", 1);
}

void FileSystem::List() {
    Stats();
    m_RootDirectory->List(0, FALSE);
}

void FileSystem::Print() {
  Stats();
  
  gBitMap->Print();
  
  printf("Root Directory:\n");
  m_RootDirectory->List(0, TRUE);
}

void FileSystem::Stats() {
  int Avail = gBitMap->NumClear();
  int Used = NumSectors - Avail;
  int PctUsed = (Used * 100) / NumSectors;
  int PctAvail = (Avail * 100) / NumSectors;
  printf("File System: %d blocks avail (%d%%), %d blocks used (%d%%)\n",
	 Avail, PctAvail, Used, PctUsed);
}

bool FileSystem::Init() {
  printf("Loading user programs into file system...\n");
  printf("Loading /bin/shell... ");
  ASSERT(Copy("../test/SUN/shell","/bin/shell") == TRUE);
  printf("Done.\nLoading /bin/exit... ");
  ASSERT(Copy("../test/SUN/exit","/bin/exit") == TRUE);
  printf("Done.\nLoading /bin/halt... ");
  ASSERT(Copy("../test/SUN/halt","/bin/halt") == TRUE);
  printf("Done.\nLoading /bin/ls... ");
  ASSERT(Copy("../test/SUN/ls","/bin/ls") == TRUE);
  printf("Done.\nLoading /usr/bin/filetest... ");
  ASSERT(Copy("../test/SUN/filetest","/usr/bin/filetest") == TRUE);
  printf("Done.\nLoading /usr/bin/cntln... ");
  ASSERT(Copy("../test/SUN/halt","/usr/bin/cntln") == TRUE);
  printf("Done.\nLoading /usr/bin/vmtest... ");
  ASSERT(Copy("../test/SUN/vmtest","/usr/bin/vmtest") == TRUE);
  printf("Done.\nLoading /var/tmp/foo.txt... ");
  if(Copy("../test/foo.txt","/var/tmp/foo.txt") != -1) printf("Done.\n");
  return TRUE;
}

/* Copy a file from UNIX to Nachos */

bool FileSystem::Copy(char *from, char *to) {
  FILE *fp;
  OpenFile* openFile;
  int amountRead, fileLength;
  char data[TransferSize];
  
  /* Open the UNIX file */

  if ((fp = fopen(from, "r")) == NULL) {       
    printf("\nCopy: couldn't open input file %s\n", from);
    return FALSE;
  }

  /* determine its length */

  fseek(fp, 0, 2);            
  fileLength = ftell(fp);
  fseek(fp, 0, 0);

  /* create the Nachos version */

  openFile = kernel->fileSystem->Create(NULL, to);
  if (openFile == NULL) {
    printf("\nCopy: couldn't create output file %s\n", to);
    return FALSE;
  }

  /* copy the bits */

  while ((amountRead = fread(data, sizeof(char), TransferSize, fp)) > 0)
    openFile->Write(data, amountRead);      
  
  delete openFile;
  fclose(fp);

  return TRUE;
}
