// openfile.cc -- routines to manage open files
// ----------------------------------------------------------------------------
// NOTE: All code is changed, so we didn't use #ifdef CHANGED in this file.
// comments at the top blown away
// they were not in sync with the new code anyway
// ----------------------------------------------------------------------------


#include "filehdr.h"
#include "openfile.h"
#include "disk.h"
#include "synchdisk.h"
#include "diskcache.h"
#include "diskbitmap.h"

extern DiskBitMap *gBitMap;
extern "C" void bzero(void *, unsigned int);
extern "C" void bcopy(const void *, void *, unsigned int);

// Open a file
OpenFile::OpenFile(int sector, bool noLimit)
{ 
    m_Hdr = FileHeader::Get(sector, noLimit);
    m_SeekPosition = 0;
    m_Busy = FALSE;
    m_Sequential = FALSE;
}

// Close a file
OpenFile::~OpenFile()
{
    if (m_Hdr != NULL) {
	// File is locked here, but unlock in Put(), for atomicity.
	FileWriteLock();
	if (m_Busy) {
	    ASSERT(m_Hdr->IsBusy());
	    m_Hdr->ResetBusy();
	}
	FileWriteUnlock();
	FileHeader::Put(m_Hdr);
    }
}

// Read/write bytes from a file, starting from m_SeekPosition
int OpenFile::Read(char *into, int numBytes)
{
   int result = ReadAt(into, numBytes, m_SeekPosition, m_Sequential);
   if (result > 0)
       m_SeekPosition += result;
   m_Sequential = TRUE;
   return (result);
}

int OpenFile::Write(char *into, int numBytes)
{
   int result = WriteAt(into, numBytes, m_SeekPosition, m_Sequential);
   if (result > 0)
       m_SeekPosition += result;
   m_Sequential = TRUE;
   return (result);
}

// Read "numBytes" starting at "position" in a file.
int OpenFile::ReadAt(char *into, int numBytes, int position, bool isSequential)
{
    ASSERT(m_Hdr != 0);
    if (numBytes <= 0) return 0; 	// check request
    FileReadLock();
    int length = m_Hdr->FileLength();
    if (position >= length) {
	FileReadUnlock();
	return ((position - length) == 0 ? 0 : -1);
    }
    if ((position + numBytes) > length)		
	numBytes = length - position;


    int lOffset = position;
    int transferCount = numBytes;
    int sector = -1, nextSector = -1;
    while (transferCount > 0) {
        int sectorOffset = lOffset % SectorSize;
        int resid = SectorSize - sectorOffset;
        if (transferCount < resid)
            resid = transferCount;
	m_Hdr->MapSector(lOffset / SectorSize, &sector, &nextSector);
        if (sector != -1) { /* sector is allocated */
	    if (sector < 0 || sector >= NumSectors || !gBitMap->Test(sector))
		printf("OpenFile:ReadAt: sector=%d, %d\n",
		       sector, gBitMap->Test(sector));
	    ASSERT(sector >= 0 && sector < NumSectors);
	    ASSERT(gBitMap->Test(sector));
            char *sect = kernel->diskCache->Read(sector);
	    bcopy(&sect[sectorOffset], into, resid);
	    if (sectorOffset + resid == SectorSize)
		// block fully read -- not likely to be used again.
		kernel->diskCache->Release(sect, FALSE);
	    else
		kernel->diskCache->Release(sect, TRUE);
        } else { /* there is a hole here -- treat it as zeros */
	    bzero(into, resid);
        }
        lOffset += resid;
        into += resid;
        transferCount -= resid;
        ASSERT((lOffset % SectorSize == 0) || (lOffset == position + numBytes));
    }
    ASSERT(lOffset == position + numBytes);
    ASSERT(transferCount == 0);
    FileReadUnlock();
    return numBytes;
}

int OpenFile::WriteAt(char *from, int numBytes, int position, bool isSequential)
{
    ASSERT(m_Hdr != 0);
    // note that we must check whether the file is busy at header level.
    // someone else might have set the file as busy.
    if (numBytes <= 0) return 0;	// check request
    if (position + numBytes > MaxFileSize)
	return (0);
    FileWriteLock();
    int length = m_Hdr->FileLength();


    int written = 0;
    int lOffset = position;
    int transferCount = numBytes;
    while (transferCount > 0) {
        int sectorOffset = lOffset % SectorSize;
        int resid = SectorSize - sectorOffset;
        if (transferCount < resid)
            resid = transferCount;
	int sector = m_Hdr->MapSectorWrite(lOffset + resid);
        if (sector == -1) {
	    FileWriteUnlock();
            printf("OpenFile: write error: no space on disk\n");
            return (written);
        }
	if (sector < 0 || sector >= NumSectors || !gBitMap->Test(sector))
	    printf("OpenFile:ReadAt: sector=%d, %d\n",
		   sector, gBitMap->Test(sector));
	ASSERT(sector >= 0 && sector < NumSectors);
	ASSERT(gBitMap->Test(sector));
        char *sect = kernel->diskCache->Read(sector);
	bcopy(from, &sect[sectorOffset], resid);
	if ((sectorOffset + resid == SectorSize) && isSequential)
	    // block fully written -- not likely to be used again.
	    kernel->diskCache->WriteAsync(sect);
	else
	    kernel->diskCache->WriteDelayed(sect, TRUE);
        lOffset += resid;
        from += resid;
        written += resid;
        transferCount -= resid;
        ASSERT((lOffset % SectorSize == 0) || (lOffset == position + numBytes));
    }
    ASSERT(lOffset == position + numBytes);
    ASSERT(written == numBytes);
    ASSERT(transferCount == 0);
    FileWriteUnlock();
    return numBytes;
}

bool OpenFile::IsDir() 
{ 
    ASSERT(m_Hdr != 0);
    return m_Hdr->IsDir(); 
}

void OpenFile::Stat(filestat_t *stat)
{
    ASSERT(m_Hdr != 0);
    FileReadLock();
    stat->fileSize = m_Hdr->FileLength();
    stat->fileBlocks = m_Hdr->FileBlocks();
    if (IsDir())
	stat->fileType = FSTAT_TYPE_DIR;
    else
	stat->fileType = FSTAT_TYPE_REG;
    stat->fileSector = m_Hdr->GetSector();
    FileReadUnlock();
}

int OpenFile::Length() 
{ 
    ASSERT(m_Hdr != 0);
    return m_Hdr->FileLength(); 
}

bool OpenFile::Truncate()
{
    ASSERT(m_Hdr != 0);
    FileWriteLock();
    bool result = m_Hdr->Truncate();
    FileWriteUnlock();
    return (result);
}

void OpenFile::SetBusy()
{
    ASSERT(m_Hdr != 0);
    ASSERT(m_Busy == FALSE);
    m_Busy = TRUE;
    FileWriteLock();
    m_Hdr->SetBusy();
    FileWriteUnlock();
}

void OpenFile::MarkDeleted()
{
    m_Hdr->MarkDeleted();
}

bool OpenFile::IsDeleted()
{
    return (m_Hdr->IsDeleted());
}

bool OpenFile::IsBusy()
{
    return (m_Hdr->IsBusy());
}

int OpenFile::GetSector()
{
    return (m_Hdr->GetSector());
}

void OpenFile::FileReadLock() { m_Hdr->FileReadLock(); }
void OpenFile::FileReadUnlock() { m_Hdr->FileReadUnlock(); }
void OpenFile::FileWriteLock() { m_Hdr->FileWriteLock(); }
void OpenFile::FileWriteUnlock() { m_Hdr->FileWriteUnlock(); }
void OpenFile::DirReadLock() { m_Hdr->DirReadLock(); }
void OpenFile::DirReadUnlock() { m_Hdr->DirReadUnlock(); }
void OpenFile::DirWriteLock() { m_Hdr->DirWriteLock(); }
void OpenFile::DirWriteUnlock() { m_Hdr->DirWriteUnlock(); }
