// addrspace.cc 
//	Routines to manage address spaces (executing user programs).
//
//	In order to run a user program, you must:
//
//	1. link with the -n -T 0 option 
//	2. run coff2noff to convert the object file to Nachos format
//		(Nachos object code format is essentially just a simpler
//		version of the UNIX executable object code format)
//	3. load the NOFF file into the Nachos file system
//		(if you haven't implemented the file system yet, you
//		don't need to do this last step)
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.
//

#include "copyright.h"
#include "main.h"
#include "addrspace.h"
#include "machine.h"
#include "noff.h"

extern "C" void bzero(void *, unsigned int);
extern "C" void bcopy(const void *, void *, unsigned int);

unsigned int AddrSpace::nextAsid = 1;

#ifdef P3_3

/* Helper Functions for addrspace */

static int MMAPEntryKeyFunc(MMAPEntry *mentry) { 
  return mentry->m_StartPage;
}

static unsigned int MMAPEntryHashFunc(int key) { 
  return (unsigned int) key;
}

#endif /* P3_3 */

//----------------------------------------------------------------------
// SwapHeader
// 	Do little endian to big endian conversion on the bytes in the 
//	object file header, in case the file was generated on a little
//	endian machine, and we're now running on a big endian machine.
//----------------------------------------------------------------------

#ifndef SIMOS
static void 
SwapHeader (NoffHeader *noffH)
{
    noffH->noffMagic = WordToHost(noffH->noffMagic);
    noffH->code.size = WordToHost(noffH->code.size);
    noffH->code.virtualAddr = WordToHost(noffH->code.virtualAddr);
    noffH->code.inFileAddr = WordToHost(noffH->code.inFileAddr);
    noffH->initData.size = WordToHost(noffH->initData.size);
    noffH->initData.virtualAddr = WordToHost(noffH->initData.virtualAddr);
    noffH->initData.inFileAddr = WordToHost(noffH->initData.inFileAddr);
    noffH->uninitData.size = WordToHost(noffH->uninitData.size);
    noffH->uninitData.virtualAddr = WordToHost(noffH->uninitData.virtualAddr);
    noffH->uninitData.inFileAddr = WordToHost(noffH->uninitData.inFileAddr);
}
#endif

//----------------------------------------------------------------------
// AddrSpace::AddrSpace
// 	Create an address space to run a user program.
//	Set up the translation from program memory to physical 
//	memory.  For now, this is really simple (1:1), since we are
//	only uniprogramming, and we have a single unsegmented page table
//----------------------------------------------------------------------

#ifdef P3_2

AddrSpace::AddrSpace(int pid) {
  owner_pid = pid;

#else

AddrSpace::AddrSpace() {

#endif / * P3_3 */


#ifndef P2_2

    pageTable = new TranslationEntry[NumPhysPages];
    for (int i = 0; i < NumPhysPages; i++) {
        pageTable[i].valid = FALSE;
    }

#endif /* ! P2_2 */

    asid = nextAsid++;

#ifdef P2_2

    m_PageTable = NULL;

#else

    // zero out the entire address space
    bzero(kernel->machine->mainMemory, MemorySize);

#endif /* P2_2 */

#ifdef P3_3
    m_MMAPEntries = new HashTable<int,MMAPEntry *>(MMAPEntryKeyFunc,
						   MMAPEntryHashFunc);
#endif /* P3_3 */

}

//----------------------------------------------------------------------
// AddrSpace::~AddrSpace
// 	Dealloate an address space.
//----------------------------------------------------------------------

AddrSpace::~AddrSpace()
{
#ifdef P2_2
  if (m_PageTable)
    kernel->memoryManager->ReleasePageTable(m_PageTable);

#ifdef P3_3

  HashIterator<int,MMAPEntry *> *hi =
    new HashIterator<int,MMAPEntry *>(m_MMAPEntries);
  List<int> *mlist = new List<int>;

  while (!hi->IsDone()) {
    mlist->Append(hi->Item()->m_StartPage);
    hi->Next();
  }

  while (!mlist->IsEmpty())
    delete m_MMAPEntries->Remove(mlist->RemoveFront());

  delete hi;
  delete mlist;
  delete m_MMAPEntries;

#endif /* P3_3 */

#else
   delete [] pageTable;
#endif
}


//----------------------------------------------------------------------
// AddrSpace::Load
// 	Load a user program into memory from a file.
//
//	Assumes that the page table has been initialized, and that
//	the object code file is in NOFF format.
//
//	"fileName" is the file containing the object code to load into memory
//----------------------------------------------------------------------

bool 
AddrSpace::Load(char *fileName) 
{
#ifdef P4_1
    OpenFile *executable = 
      kernel->fileSystem->Open(kernel->fileSystem->m_RootDirectory,
			       fileName);
#else
    OpenFile *executable = kernel->fileSystem->Open(fileName);
#endif
    NoffHeader noffH;
    unsigned int size;
    if (executable == NULL) {
      printf("Unable to open file %s\n", fileName);
      return FALSE;
    }

#ifdef P2_2

    int headerSize = sizeof (noffH);   /* size of the header */
    int textSize;                      /* size of the code (text
					  segement) */
    int iDataSize;                     /* size of the initialized data
					  (data segment) */
    int uDataSize;                     /* size of the uninitialized data
					  (BSS segement) */
    int stackSize = UserStackSize;     /* The min. stack size */

    strcpy(m_Name, fileName);
    DEBUG('a', ("AddrSpace: building address space %s\n", m_Name));

#endif /* P2_2 */

    executable->ReadAt((char *)&noffH, sizeof(noffH), 0);
#ifndef SIMOS
    if ((noffH.noffMagic != NOFFMAGIC) && 
		(WordToHost(noffH.noffMagic) == NOFFMAGIC))
    	SwapHeader(&noffH);
#endif
#ifdef P4_1
    if(noffH.noffMagic != NOFFMAGIC) return FALSE;
#else
    ASSERT(noffH.noffMagic == NOFFMAGIC);
#endif
#ifdef P2_2

    textSize = noffH.code.size;
    iDataSize = noffH.initData.size;
    uDataSize = noffH.uninitData.size;
    size = textSize + iDataSize + uDataSize + UserStackSize;

#else

// how big is address space?    
    size = noffH.code.size + noffH.initData.size + noffH.uninitData.size 
			+ UserStackSize;	// we need to increase the size
						// to leave room for the stack
#endif /* P2_2 */

    numPages = divRoundUp(size, PageSize);
    size = numPages * PageSize;
#ifndef P3_1
    ASSERT(numPages <= NumPhysPages);		// check we're not trying
						// to run anything too big --
						// at least until we have
						// virtual memory
#endif
#ifdef P2_2
    
    m_Size = size;
    m_TextDataBssSize = m_Size;

    /* allocate an empty page that will be kept unallocated to detect
       stack overflow. */

       m_Size += PAGE_SZ;

  /* account for stack size */

    m_StackSize = divRoundUp(stackSize,PageSize) * PageSize;
    m_Size += m_StackSize;

    DEBUG('a',("AddrSpace: textSz=%x,iDataSz=%x,uDataSz=%x,stackSz=%x\n",
	  textSize,iDataSize,uDataSize,stackSize));
    DEBUG('a',("AddrSpace: m_TextDataBssSize=%x,m_StackSize=%x,m_Size=%x\n",
	  m_TextDataBssSize,m_StackSize,m_Size));

    m_PageTable = kernel->memoryManager->AllocatePageTable(this, m_Size / PageSize);
    if (m_PageTable == NULL) return false;

#endif

    DEBUG(dbgAddr,("Initializing address space: %d, %d\n", numPages, size));

    // Set up page table for new process
    for (unsigned int i = 0; i < numPages; i++) {

#ifdef P2_2

      //P3 requires stack on the last virtual page
      //not yet done.
      //---VVN

	if(!kernel->memoryManager->AllocatePhysPage(m_PageTable, i)) 
	  return false;
#else

	pageTable[i].virtualPage = i;	// for now, virt page # = phys page #
	pageTable[i].physicalPage = i;
	pageTable[i].valid = TRUE;
	pageTable[i].use = FALSE;
	pageTable[i].dirty = FALSE;
	pageTable[i].readOnly = FALSE;  

#endif /* P2_2 */

    }

#ifdef P2_2

    // Read text + data in. Since MemWrite splits copy on a page boundary,
    // let's just read one page at a time (also we waste less space for
    // buffer).

    VirtAddr vaddr = 0;
    int fileOffset = headerSize;
    int vpnum, ppnum;     //vitual & corresponding phys page num
    char* buffer;

    buffer = new char [PageSize];
    while (vaddr < textSize + iDataSize) {
      int count = PageSize;

      /* text+data size is not a multiple of a page */

      if (textSize + iDataSize - vaddr < PageSize)
	count = textSize + iDataSize - vaddr;

      vpnum = vaddr/PageSize;
      ppnum = m_PageTable->PhysPageNum(vpnum);

      //skip the header in the the binary file
      //fileOffset points to the loc just past the header
      //after the header read a block of size PageSize
      //of course, if the binary is real small, it will be less that PageSize
      //the correct size to read is given by count
      //the output is in buffer
      /* printf("AddrSpace: loading vpn=#%x(ppn=#%x),offset=%x,count=%x\n",
	    vpnum,
	    ppnum,
	    fileOffset,
	    count); */

      executable->ReadAt(buffer, count, fileOffset);
      MemWrite(buffer, vaddr, count);
      vaddr += count;
      fileOffset += count;
    }
    delete [] buffer;

    // now vaddr points to the first byte of bss.  

    // Since we keep a 'red-zone' page to detect stack overflow, we must
    // zero out bss and stack in separate loops.

    //printf("zeroing bss, %x .. %x\n", vaddr, vaddr + uDataSize);
    if (uDataSize != 0) {

      // bss is quite small, so we zero it in one shot.

      buffer = new char [uDataSize];
      bzero(buffer, uDataSize);
      MemWrite((void *)buffer, vaddr, uDataSize);
      delete [] buffer;
    }

    DEBUG('a',("allocating/zeroing stack, %x .. %x\n",
	  size + PageSize, m_Size));

    // now we allocate and initialize each stack page.
    // stack is zeroed one page at a time.

    buffer = new char [PageSize];
    bzero(buffer, PageSize);

#ifdef P3_2

    /* for VM the address 0x7ffffffc is actually the address of the
       last word in memory, so we need to put stack on last page */

    m_Size = HIGHESTADDRESS;
    vaddr = m_Size/PageSize;

    //printf ("Zeroing Stack,Virtual Page %d\n",vaddr);
    if (!kernel->memoryManager->AllocatePhysPage(m_PageTable,vaddr)) {
      delete [] buffer ;
      return false;
    }
    
    MemWrite((void *)buffer, vaddr*PageSize, PageSize);

#else

    for (vaddr = size + PageSize; vaddr < m_Size; vaddr += PageSize) {

      /* allocate page */

      //printf ("Zeroing Stack,Virtual Page %d\n",vaddr/PageSize);
      if (!kernel->memoryManager->AllocatePhysPage(m_PageTable,vaddr/PageSize)) {
	delete [] buffer ;
	return false;
      }

      MemWrite((void *)buffer, vaddr, PageSize);

    }

#endif /* P3_2 */

    delete [] buffer;


#else

    // then, copy in the code and data segments into memory
    if (noffH.code.size > 0) {
        DEBUG(dbgAddr,("Initializing code segment."));
        DEBUG(dbgAddr,("%d, %d\n", noffH.code.virtualAddr, noffH.code.size));

        int addr = (pageTable[noffH.code.virtualAddr/PageSize].physicalPage *
                    PageSize + (noffH.code.virtualAddr % PageSize));

        executable->ReadAt(&(kernel->machine->mainMemory[addr]), 
                           noffH.code.size, noffH.code.inFileAddr);
    }

    if (noffH.initData.size > 0) {
        DEBUG(dbgAddr,("Initializing data segment."));
	DEBUG(dbgAddr,("%d, %d\n", noffH.initData.virtualAddr, noffH.initData.size));
        int addr = (PageSize * pageTable[noffH.initData.virtualAddr/PageSize].physicalPage + noffH.initData.virtualAddr % PageSize);

        executable->ReadAt(&(kernel->machine->mainMemory[addr]),
                           noffH.initData.size, noffH.initData.inFileAddr);
    }

#endif /* P2_2 */

    delete executable;			// close file
    return TRUE;			// success
}

//----------------------------------------------------------------------
// AddrSpace::Execute
// 	Run a user program.  Load the executable into memory, then
//	(for now) use our own thread to run it.
//
//	"fileName" is the file containing the object code to load into memory
//----------------------------------------------------------------------

void 
AddrSpace::Execute(char *fileName) 
{
    kernel->currentThread->space = this;

    if (!Load(fileName)) {
#ifdef P2_2_DEBUG
      printf ("ERROR: Could not load %s.\n",fileName);
#endif      
      kernel->currentThread->space = NULL;
      return;				// executable not found
    }

#ifdef P2_2_DEBUG
      printf ("DEBUG: Loaded %s.\n",fileName);
#endif      

    this->InitRegisters();		// set the initial register values
    this->RestoreState();		// load page table register

    kernel->machine->Run();		// jump to the user progam

    ASSERTNOTREACHED();			// machine->Run never returns;
					// the address space exits
					// by doing the syscall "exit"
}


//----------------------------------------------------------------------
// AddrSpace::InitRegisters
// 	Set the initial values for the user-level register set.
//
// 	We write these directly into the "machine" registers, so
//	that we can immediately jump to user code.  Note that these
//	will be saved/restored into the currentThread->userRegisters
//	when this thread is context switched out.
//----------------------------------------------------------------------

void
AddrSpace::InitRegisters()
{
    Machine *machine = kernel->machine;
    int i;

#ifdef P2_2
    int stackptr;
#endif

    for (i = 0; i < NumTotalRegs; i++)
	machine->WriteRegister(i, 0);

    // Initial program counter -- must be location of "Start"

    machine->WriteRegister(PCReg, 0);	

    // Need to also tell MIPS where next instruction is, because
    // of branch delay possibility

    machine->WriteRegister(NextPCReg, 4);

   // Set the stack register to the end of the address space, where we
   // allocated the stack; but subtract off a bit, to make sure we don't
   // accidentally reference off the end!

#ifdef P2_2

#ifndef P3_1
    stackptr = m_TextDataBssSize + 1;    /* last byte of addr space */
#else
    stackptr = m_Size - 17;
#endif /* ! P3_1 */

    stackptr &= ~3;                      /* word align */

    //printf("StackPtr=%d\n",stackptr);
    machine->WriteRegister(StackReg,stackptr);
    DEBUG(dbgAddr,("Initializing stack pointer: %d\n",stackptr));

#else

    machine->WriteRegister(StackReg, numPages * PageSize - 16);
    DEBUG(dbgAddr,("Initializing stack pointer: %d\n", numPages*PageSize - 16));

#endif /* P3_2 */
}

//----------------------------------------------------------------------
// AddrSpace::SaveState
// 	On a context switch, save any machine state, specific
//	to this address space, that needs saving.
//
//	For now, don't need to save anything!
//----------------------------------------------------------------------

void AddrSpace::SaveState() 
{

}

//----------------------------------------------------------------------
// AddrSpace::RestoreState
// 	On a context switch, restore the machine state so that
//	this address space can run.
//
//      For now, tell the machine where to find the page table.
//----------------------------------------------------------------------

void AddrSpace::RestoreState() 
{
#ifndef VM

#ifndef P2_2
    kernel->machine->pageTable = pageTable;
#endif

    kernel->machine->pageTableSize = numPages;

#else 

    kernel->machine->pageTable = NULL;
    if(kernel->machine->tlb) {
      for(int i = 0 ; i < TLBSize; i++) {
	kernel->machine->tlb[i].valid = FALSE;
	//now the first memory reference will cause a page-fault
      }
    } else {
      DEBUG(dbgAddr, ("ERROR!!! - In thread \"%s\": AddrSpace::RestoreState()\n\tDid not find TLB\n", kernel->currentThread->getName()) );
    }

#endif /* VM */

#ifdef SIMOS
    UpdateTLB(asid);
#endif
}

#ifdef P2_2

bool AddrSpace::ValidPage(int vPageNum) {
  return (m_PageTable->IsValid(vPageNum) ? TRUE : FALSE);
}

int AddrSpace::MemRead( VirtAddr addr, void *data, int length ) {

  DEBUG('a',("MemRead --> %x, #%d bytes\n", addr, length));

  // check for out of bounds address
  if (length < 0 || addr >= m_Size || (addr + length) >= m_Size)
    return (0);

  // a buffer can span across multiple virtual pages, so we must transfer
  // no more than one page at a time to use the correct physical
  // address(es).
 
  while (length > 0) {

    /* get offset within page of 'addr' */

    int vPageNum = addr / PageSize;
    int pageOffset = addr % PageSize;
    int resid = PageSize - pageOffset;

    /* adjust it if necessary */

    if (length < resid) resid = length;

    /* check if this is a valid page */

    if (!ValidPage(vPageNum)) {
      printf("MemRead: invalid memory virt page #%x\n", vPageNum);
      return (0);
    }

    int paddr = m_PageTable->PhysPageNum(vPageNum) * PageSize;
    paddr += pageOffset;
    DEBUG('a',("MemRead: reading #%d bytes into phys %x\n", resid, paddr));
    bcopy((void*)&(kernel->machine->mainMemory[paddr]), data, resid);

    addr += resid;
    data = (void *)((char *)data + resid);
    length -= resid;
  }
  return 1;
}

int 
AddrSpace::MemWrite( void *data, VirtAddr addr, int length )
{
  DEBUG('a',("MemWrite --> %x, #%d bytes\n", addr, length));
  // check for out of bounds address
  if (length < 0 || addr >= m_Size || (addr + length) >= m_Size)
        return (0);
  //
  // a buffer can span across multiple virtual pages, so we must transfer
  // no more than one page at a time to use the correct physical
  // address(es).
  //
  while (length > 0) {
    // get offset within page of 'addr'.
        int vPageNum = addr / PageSize;
        int pageOffset = addr % PageSize;
        int resid = PageSize - pageOffset;
	//hack
        resid = (resid==PageSize)?resid-1:resid;
        // adjust it if necessary.
        if (length < resid)
                resid = length;
        if (!ValidPage(vPageNum)) {
                printf("MemWrite: invalid memory virt page #%x\n",
		       vPageNum);
                return (0);
        }
        int paddr = m_PageTable->PhysPageNum(vPageNum) * PageSize;
        paddr += pageOffset;
        DEBUG('a',("MemWrite: writing #%d bytes into phys %x\n", resid,
	      paddr));
        bcopy(data, (void*)&(kernel->machine->mainMemory[paddr]), resid);

        addr += resid;
        data = (void *)((char *)data + resid);
        length -= resid;
  }
  return 1;
}

#ifdef P3_1
bool AddrSpace::VPageExistsInMainMemory(int vpn, int *ppnPtr)
{
  *ppnPtr = m_PageTable->PhysPageNum(vpn);
  return ((*ppnPtr >= 0) ? TRUE : FALSE );
}
#endif /* P3_1 */
#endif /* P2_2 */




