// addrspace.h 
//	Data structures to keep track of executing user programs 
//	(address spaces).
//
//	For now, we don't keep any information about address spaces.
//	The user level CPU state is saved and restored in the thread
//	executing the user program (see thread.h).
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.
//

#ifndef ADDRSPACE_H
#define ADDRSPACE_H

#include "copyright.h"
#include "machine.h"
#include "filesys.h"

#ifdef P3_2
#define HIGHESTADDRESS 0x7ffffffc
#endif

#ifdef P3_3
#include "hash.h"
#include "list.h"
#endif

#define UserStackSize		1024 	// increase this as necessary!

#ifdef P2_2

#define STACKTOP                96      //  SPARC convention 
#define MAXPATH                 64      //  may change
#define PAGE_SZ                 4096    //  may change

typedef int VirtAddr;
class PageTable;
class MemoryManager;

#endif /* P2_2 */

#ifdef P3_3

/* class representing a set of mmap'ed pages in memory */

class MMAPEntry {

 public:

  MMAPEntry(int fd,int startpage,int bytes) {
    m_fd = fd;
    m_StartPage = startpage;
    m_Length = bytes;
  }

  ~MMAPEntry() {};

  int m_fd;          /* the file descriptor corresponding to the file */
  int m_StartPage;   /* the starting page of the mmap'ed file, 
			same as the SegmentID returned by doMmap() */
  int m_Length;      /* length of the mmap'ed file in bytes */
};

#endif /* P3_3 */

class AddrSpace {
  public:
#ifdef P3_2
  AddrSpace(int pid);
#else
    AddrSpace();			// Create an address space.
#endif
    ~AddrSpace();			// De-allocate an address space

    void Execute(char *fileName);	// Run the the program
					// stored in the file "executable"

    void SaveState();			// Save/restore address space-specific
    void RestoreState();		// info on a context switch 
    unsigned int ASID() { return asid; }

#ifdef P3_2
    int PID() { return owner_pid; }
#endif

#ifdef P2_2

    int Size(){ return m_Size; }        // return size of address space in bytes

    // Copy length bytes from user address space to data.
    // You may want to return an error code from MemRead MemWrite.

    int MemRead( VirtAddr addr, void *data, int length );

    // Copy length bytes from data to user address space.

    int MemWrite( void *data, VirtAddr addr, int length );
#ifdef P3_1
    bool VPageExistsInMainMemory(int vpn, int *ppn);
#endif   
    friend class MemoryManager;

#endif /* P2_2 */

    // Enable ExceptionHandler to access the page table
    friend void ExceptionHandler(ExceptionType which);

  private:

#ifdef P2_2

    bool ValidPage(int vPageNum);      /* TRUE if 'vPageNum' is a valid
				          pagenum for this address
				          space. */

#endif /*P2_2*/

#ifndef P2_2
    TranslationEntry *pageTable;	// Assume linear page table translation
					// for now!
#endif

    unsigned int numPages;		// Number of pages in the virtual 
					// address space

    bool Load(char *fileName);		// Load the program into memory
					// return false if not found

    void InitRegisters();		// Initialize user-level CPU registers,
					// before jumping to user code

    static unsigned int nextAsid;       // Need ASID for MIPS TLB
    unsigned int asid;

#ifdef P3_2
    int owner_pid;
#endif

#ifdef P2_2

    int  m_StackSize;                   /* size of stack (page aligned) */
    int  m_Size;                        /* size of address space in bytes */
    int  m_TextDataBssSize;             /* size of text + data + bss
					   (page aligned) */

 public:
    char m_Name[MAXPATH + 1];           /* name of executable image */
    PageTable *m_PageTable;

#ifdef P3_3    
    HashTable<int,MMAPEntry *> *m_MMAPEntries;   /* store a table of
						   the mmaped pages in
						   this addr space */
#endif /* P3_3 */

#endif /* P2_2 */

};

#endif // ADDRSPACE_H

