// exception.cc 
//	Entry point into the Nachos kernel from user programs.
//	There are two kinds of things that can cause control to
//	transfer back to here from user code:
//
//	syscall -- The user code explicitly requests to call a procedure
//	in the Nachos kernel.  Right now, the only function we support is
//	"Halt".
//
//	exceptions -- The user code does something that the CPU can't handle.
//	For instance, accessing memory that doesn't exist, arithmetic errors,
//	etc.  
//
//	Interrupts (which can also cause control to transfer from user
//	code into the Nachos kernel) are handled elsewhere.
//
// For now, this only handles the Halt() system call.
// Everything else core dumps.
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.
//

#include "copyright.h"
#include "main.h"
#include "syscall.h"
#include "fileio.h"

#ifdef P3_3
#include "addrspace.h"
#endif

#ifdef P2_2

#include "registers.h"
#include "utility.h"

#define MAXARGC  16      /* max number of args */
#define MAXARGL 512      /* max length of each arg in argc */
#define STACKTP  96      /* top of the sparc stack */
#define FAILURE  -1      

/* helper function to handle Exit() */

static void doExit(int code) {

#ifdef P2_2_DEBUG
  printf("Exit(): Called by \"%s\" with arg = %d\n", 
	 kernel->currentThread->getName(), code);
#endif

  /* close all files open by this proc */

  delete kernel->currentThread->fileio;
  kernel->currentThread->fileio = NULL;

  /* inform the PidManager that this proc has terminted */

  kernel->pidManager->PidTerminate(kernel->currentThread->pid);

  /* inform the Scheduler that this proc has finished */

  kernel->currentThread->Finish();
}

/* helper function to exec a process in its own thread */

static void RunProcess(char *prog) {

#ifdef P2_2_DEBUG
  printf("Executing process %s.\n",
	 kernel->currentThread->space->m_Name);
#endif

  int this_pid = kernel->currentThread->pid;
  kernel->currentThread->space->Execute(kernel->currentThread->space->m_Name);
  doExit(this_pid);
}

/* helper function to handle Join() */

static int doJoin(int pid) {
#ifdef P2_2_DEBUG
  printf("Join(): Called by %s on PID %d.\n",
	 kernel->currentThread->getName(), pid);
#endif
  return ((pid >= 0) ?
	  kernel->pidManager->PidWait(pid) : FAILURE);
}

/* helper function handle Exec() */

int doExec(int arg1,int arg2, int arg3) {

  char cmdline[256];
  int i, argc, arg_size;
  char *argv[MAXARGC];
  char *cur_arg,*prog;

  /* get the command line */

  kernel->memoryManager->CopyStr(cmdline,
				 kernel->currentThread->space,
				 arg1);

  cur_arg = strtok(cmdline," ");
  prog = ((!cur_arg) ? cmdline : cur_arg);

  if (prog[0] == '\0') return FAILURE;

#ifdef P2_2
  printf("Exec(): Parent (%s) Child (%s).\n", 
	 kernel->currentThread->getName(),
	 prog);
#endif

#ifdef P2_3_FOO

  /* initialize argc,argv */

  argc=0;
  for (i = 0; i < MAXARGC ; i++) argv[i]=0;

  if (argc > MAXARGC) {
    printf("Exec(): Invalid number of args %d\n",argc);
    return FAILURE;
  }

  for (i = 0; i < argc ; i++)
      printf("Exec(): argv[%d] = %s.\n",i,argv[i]);

#endif /* P2_3 */

  /* everything looks okay, do the exec */

#ifdef P3_1

  /* create a separate thread for this process */

  Thread *new_th = new Thread(prog, TRUE);
  if (new_th == NULL) {
    printf("Exec(): Cannot allocate a thread.\n");
    delete new_th;
    return FAILURE;
  }

  /* get a PID for the process */

  PID new_pid = kernel->pidManager->PidAllocate(new_th);
  if (new_pid == NOPID) {
    printf("Exec(): Could not allocated a pid.\n");
    delete new_th;
    return FAILURE;
  }

  /* allocate new file io for this process */

#ifdef P4_1
  Directory *clonedDir;
  clonedDir = kernel->fileSystem->CloneDir(kernel->currentThread->fileio->GetCurrentDir());
  FileIO *new_fileio = new FileIO(clonedDir);
#else
  FileIO *new_fileio = new FileIO;
#endif

  if (new_fileio == NULL) {
    printf("Exec(): Cannot allocate file I/O.\n");
    delete new_fileio;
    delete new_th;
    kernel->pidManager->PidTerminate(new_pid);
    return FAILURE;
  }

#else

  /* get a PID for the process */

  PID new_pid = kernel->pidManager->PidAllocate();
  if (new_pid == NOPID) {
    printf("Exec(): Could not allocated a pid.\n");
    return FAILURE;
  }

  /* allocate new file io for this process */

  FileIO *new_fileio = new FileIO;
  if (new_fileio == NULL) {
    printf("Exec(): Cannot allocate file I/O.\n");
    delete new_fileio;
    kernel->pidManager->PidTerminate(new_pid);
    return FAILURE;
  }

  /* create a separate thread for this process */

  Thread *new_th = new Thread(prog, TRUE);
  if (new_th == NULL) {
    printf("Exec(): Cannot allocate a thread.\n");
    delete new_th;
    delete new_fileio;
    kernel->pidManager->PidTerminate(new_pid);
    return FAILURE;
  }

#endif /* P3_1 */

  /* create a new address space for the process */

#ifdef P3_2
  AddrSpace *new_as = new AddrSpace(new_pid);
#else
  AddrSpace *new_as = new AddrSpace();
#endif

  strcpy(new_as->m_Name,prog);

  /* all the allocations were successful, load the prog into
     its new address space */

  new_th->pid    = new_pid;
  new_th->space  = new_as;
  new_th->fileio = new_fileio;

  printf("Exec(): Current PID=%d, New PID=%d.\n",
	 kernel->currentThread->pid,new_th->pid);

  kernel->currentThread->SaveUserState();
  new_th->Fork((VoidFunctionPtr)RunProcess,(void *)prog);
  return new_pid;
}

#endif /* P2_2 */

#ifdef P3_3

/* function to handle mmap'ing a file */

static int doMmap(int fd,int vaddr,int bytes) {

  int vpn_start, vpn_end;      /* starting and ending virt pages */
  int addr_end;                /* The ending address, save it in vars,
				  to avoid doing extra math in
				  loop. This value is required to
				  make sure the last memwrite does not
				  address a invalid loc in the file */
  int count;
  MMAPEntry *cur_mmapfile;

  FileIO* fileio = kernel->currentThread->fileio;
  AddrSpace *cur_as = kernel->currentThread->space;
  PageTable *cur_pt = cur_as->m_PageTable;

  /* if the vaddr is outside bounds, return failure */

  if (vaddr < 0 || vaddr > cur_as->Size()) return FAILURE;

  addr_end = vaddr + bytes;     

  /* make sure stack is not overwritten */

  if (addr_end > (cur_as->Size() - PageSize)) return FAILURE;

  if (addr_end > HIGHESTADDRESS) return FAILURE;

  vpn_start = vaddr/PageSize;           
  vpn_end = (vaddr + bytes)/PageSize;

#ifdef P3_3_DEBUG  
  printf("MMAP(): Ready To Allocate Pages, %d to %d.\n",vpn_start,vpn_end);
#endif

  int i;
  for (i=vpn_start;i<=vpn_end;i++) {

    /* if vaddr > addr_end, finsihed with mmap'ing */

    if (vaddr >= addr_end) break;

    /* make sure not mmap'in already mmap'ed page */

    if (cur_as->m_MMAPEntries->IsInTable(i)) {
      printf("MMAP(): Page %d is already mmap'ed\n.",i);
      return FAILURE;
    }

    /* request pages from memory manager */

    if (!kernel->memoryManager->AllocatePhysPage(cur_pt,i)) {
      printf("MMAP(): Could not allocate page %d\n.",i);
      return FAILURE;
    }

#ifdef P3_3_DEBUG  
    printf("MMAP(): Allocated Page, %d.\n",i);
#endif

    /* determine how many bytes to read from file, read them and write
       them to memory */

    count = fileio->FileRead(fd,(char *)vaddr,PageSize-1);
    if (count == FAILURE) { 
      printf("MMAP(): FileRead from file %d failed.\n",fd);
      return FAILURE;
    }
#ifdef P3_3_DEBUG  
    printf("MMAP(): Wrote %d bytes from file %d to memory.\n",count,fd);
#endif
    vaddr += count;
  }

#ifdef P3_3_DEBUG  
  printf("MMAP(): Done Allocating Pages, %d to %d.\n",vpn_start,vpn_end);
#endif

  /* we were able to mmap the file sucessfully, so enter these pages
     into the address spaces table of mmap'ed pages */

  cur_mmapfile = new MMAPEntry(fd,vpn_start,bytes);
  cur_as->m_MMAPEntries->Insert(cur_mmapfile);

#ifdef P3_3_DEBUG  
  printf("MMAP(): Added %d to hashtable.\n",vpn_start);
#endif

  /* return the starting vpn of the mmap'ed file as the SID */

  return vpn_start; 
}

/* function to handle unmmap'ing a file */

static void doMunmap(int SID) {

  MMAPEntry *to_unmap;
  int vpn_start,vpn_end,vaddr,addr_end,count;

  FileIO* fileio = kernel->currentThread->fileio;
  AddrSpace *cur_as = kernel->currentThread->space;

  /* if the given SID is not in the hash table return */

  if (!cur_as->m_MMAPEntries->Find(SID,&to_unmap)) {
    printf("MUNMAP(): SID %d not mmap'ed.\n",SID);
    return;
  }

  vpn_start = to_unmap->m_StartPage;  
  vaddr = vpn_start * PageSize;
  addr_end = vaddr + to_unmap->m_Length;
  vpn_end = addr_end/PageSize;

  /* write pages back, starting at the beginning of the file,
   * otherwise file grows on disk */

  fileio->FileSeek(to_unmap->m_fd,0);

  for (int i=vpn_start;i<=vpn_end;i++) {

    /* determine the number of bytes to read from memory
       and write to the file */

    count = fileio->FileWrite(to_unmap->m_fd,(char *)vaddr,PageSize-1);
    kernel->memoryManager->ReleasePage(cur_as->m_PageTable,i);
    if (count == FAILURE) break;
    vaddr += count;
  }

  delete cur_as->m_MMAPEntries->Remove(SID);

#ifdef P3_3_DEBUG  
  printf("MUNMAP(): mmap'ed file %d unmmap'ed.\n",SID);
#endif
}

#endif /* P3_3 */

//----------------------------------------------------------------------
// ExceptionHandler
// 	Entry point into the Nachos kernel.  Called when a user program
//	is executing, and either does a syscall, or generates an addressing
//	or arithmetic exception.
//
// 	For system calls, the following is the calling convention:
//
// 	system call code -- r2
//		arg1 -- r4
//		arg2 -- r5
//		arg3 -- r6
//		arg4 -- r7
//
//	The result of the system call, if any, must be put back into r2. 
//
// And don't forget to increment the pc before returning. (Or else you'll
// loop making the same system call forever!
//
//	"which" is the kind of exception.  The list of possible exceptions 
//	are in machine.h.
//----------------------------------------------------------------------

void
ExceptionHandler(ExceptionType which)
{
    int type   = kernel->machine->ReadRegister(2);

#ifdef P2_2

    int term_pid;

    /* get the parameters passed to the the system call */

    int param1 = kernel->machine->ReadRegister(4);
    int param2 = kernel->machine->ReadRegister(5);
    int param3 = kernel->machine->ReadRegister(6);
    int param4 = kernel->machine->ReadRegister(7);

#ifdef P3_1 
    int vpn;
    TEWithPid te;
#endif

#else

    int retval = kernel->machine->ReadRegister(4);

#endif /* P2_2 */


    FileSystem* fs = kernel->fileSystem;

#ifndef P2_2
    int addr;
    bool result;
#endif

    AddrSpace* asptr = kernel->currentThread->space;
    char str[256];      
    int npc;

#ifdef P2_2
    FileIO* fileio = kernel->currentThread->fileio;
    OpenFileId fd;
    int return_code = FAILURE;  /* assume all syscalls fail */
    Thread * tmp_th;
#endif /* P2_2 */

#ifdef P4_1
    int msg_len;
#endif

    switch (which) {

#ifdef P3_1

    case PageFaultException:

      vpn = kernel->machine->ReadRegister(BadVAddrReg);
      vpn = vpn/4096;
      term_pid = kernel->currentThread->pid;
      //printf("PageFaultException for PID %d - VA 0x%x\n", term_pid, vpn);
      if (kernel->memoryManager->HandleTlbMiss(term_pid, vpn, &te)) return;      

      /* if the memory manager couldn't handle the TlbMiss, kill the
	 current proc */

      doExit(term_pid);
      return;
      break;

#endif /* P3_1 */

    case SyscallException:

      switch(type) {

      case SC_Halt:
	DEBUG(dbgAddr, ("Shutdown, initiated by user program.\n"));
	kernel->fileSystem->Sync();
	kernel->interrupt->Halt();
	break;

#ifdef P2_2

	/* Exit() System Call, exits the current process with a given
	   return code */

      case SC_Exit:

#ifdef P3_2
	doExit(param1);
#else
	tmp_th = kernel->currentThread;
	doExit(param1);
	delete tmp_th->space;
	tmp_th->space = NULL;
#endif /* P3_2 */

	break;

	/* Create() System Call, creates the named file, return
	   value indicates sucess (1) or failure (0) */

      case SC_Create:

	// param1 is user-space virt-addr

	kernel->memoryManager->CopyStr(str, asptr, param1);
        if(str && str[0] != '\0')
#ifdef P4_1

	  kernel->machine->WriteRegister(2,fileio->FileCreate((char*)str));
#else
	  kernel->machine->WriteRegister(2,fs->Create((char*)str));
#endif
	break;

	/* Open() System Call, opens the named file, retrun value is
	   the file descriptor that was opened */

      case SC_Open:

	kernel->memoryManager->CopyStr(str, asptr, param1);
	fd = fileio->FileOpen((char*)&str[0]);
	kernel->machine->WriteRegister(2,fd);      
	break;

	/* Close() System Call, closes the given file descriptor, always
	   returns sucess (1) */

      case SC_Close:

	(void)fileio->FileClose(param1);
	break;

	/* Read() System Call, reads bytes from the named file, return
	   value is the number of bytes read */

      case SC_Read:

	return_code = fileio->FileRead(param3, (char*)param1, param2);
	kernel->machine->WriteRegister(2,return_code);	
	break;

	/* Write() System Call, writes bytes to the named file, return
           value is the number of bytes written */

      case SC_Write:

	return_code = fileio->FileWrite(param3, (char*)param1, param2);
	fflush(stdout);
	kernel->machine->WriteRegister(2,return_code);	
	break;

	/* Join() System Call, allows one process to wait on another,
	   return code is the exit status of the pid being waited on */

      case SC_Join:

	kernel->machine->WriteRegister(2,doJoin(param1));
	break;

	/* Exec() System Call, executes a new process */

      case SC_Exec:
	
	kernel->machine->WriteRegister(2,doExec(param1,param2,param3));
	break;

#endif /* P2_2 */

#ifdef P3_3

	/* Length() System Call, returns the length of an open file */

      case SC_Length:

	kernel->machine->WriteRegister(2,fileio->FileLength(param1));
	break;

	/* Mmap() System Call, maps the given file onto the given virt addr */

      case SC_Mmap:

	kernel->machine->WriteRegister(2,doMmap(param1,param2,param3));
        break;

	/* Munmap() System Call, unmaps the given file from memory */

      case SC_Munmap:

	doMunmap(param1);
	break;

#endif /* P3_3 */

#ifdef P4_1

	/* Chdir() System Call, changes the current working directory */

      case SC_Chdir:

	kernel->memoryManager->CopyStr(str, asptr, param1);
	//printf("Chdir %s called!\n",str);
	kernel->machine->WriteRegister(2,fileio->ChDir(str));
	break;

	/* Lsdir() System Call, lists the contents of the current directory */

      case SC_Lsdir:		

	fileio->GetCurrentDir()->List(0,FALSE);
	break;

	/* Mkdir() System Call, makes a new directory */

      case SC_Mkdir:

	kernel->memoryManager->CopyStr(str, asptr, param1);
	kernel->machine->WriteRegister(2,fileio->MkDir(str));
	break;

	/* Remove() System Call, removes a file */

      case SC_Remove:

	kernel->memoryManager->CopyStr(str, asptr, param1);
	fileio->FileRemove(str);
	break;

	/* Print() System Call, prints a string to STDOUT */

      case SC_Print:

	kernel->memoryManager->CopyStr(str, asptr, param1);
	msg_len = strlen(str);
	for (int z=0;z<msg_len;z++) kernel->consoleOutput->PutChar(str[z]);
	fflush(stdout);
	break;

	/* Readch() System Call, reads a string to STDIN */

      case SC_Readch:

	kernel->machine->WriteRegister(2,kernel->consoleInput->GetChar());
	break;

      case SC_Pwd:

	strcpy(str,fileio->GetPwd());
	msg_len = strlen(str);
        for (int z=0;z<msg_len;z++) kernel->consoleOutput->PutChar(str[z]);
	kernel->consoleOutput->PutChar('\n');
	break;
	
#endif /* P4_1 */

      default:
	printf("Unexpected system call %d\n", type);
	tmp_th = kernel->currentThread;
	doExit(tmp_th->pid);
#ifndef P3_2
	delete tmp_th->space;
	tmp_th->space = NULL;
#endif
	break;
      }
      
#ifdef P2_2

      /* update PC registers */

      npc = kernel->machine->ReadRegister(NextPCReg);
      kernel->machine->WriteRegister(PrevPCReg,
				     kernel->machine->ReadRegister(PCReg));
      kernel->machine->WriteRegister(PCReg, npc);
      kernel->machine->WriteRegister(NextPCReg, npc + 4);	  
      return;

#endif /* P2_2 */

      break;


    default:
      printf("Unexpected user mode exception %d\n", which);
      tmp_th = kernel->currentThread;
      doExit(tmp_th->pid);
#ifndef P3_2
      delete tmp_th->space;
      tmp_th->space = NULL;
#endif
      return;
      break;
    }
    ASSERTNOTREACHED();
}



