// fileio.cc - per user i/o calls

#include "fileio.h"
#include "addrspace.h"
#include "thread.h"
#include "utility.h"

#ifdef P4_1
FileIO::FileIO(Directory *cDir)
#else
FileIO::FileIO()
#endif
{

#ifdef P4_1
    currentDir = cDir;
    strcpy(pwd,"/");
#endif

    int i;
    for (i = 0; i < NFILES; i++) {
        openFileSlot[i].fd = -1;
        openFileSlot[i].name = NULL;
        openFileSlot[i].file = NULL;
    }

#ifndef P4_1
    i = 0;
    openFileSlot[i].fd = i;
    openFileSlot[i].name = "Stdin";
    openFileSlot[i].file = new OpenFile(i);

    i = 1;
    openFileSlot[i].fd = i;
    openFileSlot[i].name = "Stdout";
    openFileSlot[i].file = new OpenFile(i);
#endif /* P4_1 */

}

// close all open files
FileIO::~FileIO() {

  for (int i = 0; i < NFILES; i++) {
    if (openFileSlot[i].file != NULL) {
      if (openFileSlot[i].fd != 0 && 
	  openFileSlot[i].fd != 1) {
	(void)FileClose(openFileSlot[i].fd);
      }
    }
  }
#ifdef P4_1
  delete currentDir;
#endif
}

/*****************************************************
if you open an already open file, FileClose will not know which one to close!!!
Assume that the file has already been created
*****************************************************/
OpenFileId FileIO::FileOpen(char *filename)
{
    // find an unused slot
   int i;

   if(!filename || 0==strlen(filename))return (-1);

    for (i = 0; i < NFILES; i++) {
      if (openFileSlot[i].file == NULL)
	break;
    }

    if (i == NFILES) { // too many open files
        printf("Open %s: too many open files\n", filename);
        return (-1);
    }
    struct OpenFileSlot *slot = &openFileSlot[i];

    // Try to open the file

#ifdef P4_1
    slot->file = kernel->fileSystem->Open(currentDir,filename);
#else
    slot->file = kernel->fileSystem->Open(filename);
#endif

    if (slot->file == NULL) {
      printf("FileIO::FileOpen(\"%s\") :- Failed in FileSystem::Open(...)\n", 
	     filename);
      return -1;
    }

    slot->name = new char [strlen(filename) + 1];
    strcpy(slot->name, filename);
#ifndef P4_1
    slot->fd = slot->file->file;
#endif
    DEBUG('a',("FileIO: Open %s: ID=%d\n", filename, slot->fd));
    return (slot->fd);
}

bool FileIO::FileClose(OpenFileId fd) {

    if (fd < 0 ) {
        printf("Close %d: invalid ID\n", fd);
        return (FALSE);
    }
    struct OpenFileSlot *slot;
    for(int i = 0; i < NFILES; i++){
      slot = &openFileSlot[i];
      if(slot->fd == fd) {
        delete slot->file;
        slot->file = NULL;
        delete [] slot->name;
        slot->name = NULL;
	slot->fd = -1;
	return TRUE;
      }
    }
    return FALSE;
}

int FileIO::FileWrite(OpenFileId fd, char *bp, int nbytes)
{
  if(fd<0)return -1;
  if (nbytes == 0) return (0);
  if (nbytes < 0 || nbytes > MAX_IO_SIZE) return (-1);
  struct OpenFileSlot *slot;
  for(int i = 0; i < NFILES; i++){
    slot = &openFileSlot[i];
    if (slot->file->IsDir()) {
      printf("Write %s: cannot write on a directory\n", slot->name);
      return (-1);
    }
    if (slot->file->IsBusy()) {
        printf("Write %s: paging file busy\n", slot->name);
        return (-1);
    }
    if(slot->fd == fd) {
      char *buffer = new char [nbytes];
      if(!kernel->memoryManager->ReadBytes(buffer, 
					   nbytes, 
					   kernel->currentThread->space, 
					   (int)bp)) {
	delete [] buffer;
	return (-1);
      }
      int ret = slot->file->Write(buffer, nbytes);
      delete [] buffer;
      return (ret);
    }
  }
  return -1;
}

/************************************
read nbytes from the file and copy them to virtaddr bp
************************************/
int FileIO::FileRead(OpenFileId fd, char *bp, int nbytes)
{
 
  if(fd<0)return -1;
  if (nbytes == 0) return (0);
  if (nbytes < 0 ) return (-1);
  struct OpenFileSlot *slot;
  for(int i = 0; i < NFILES; i++){
    slot = &openFileSlot[i];
    if(slot->fd == fd) {
      char *buffer = new char [nbytes];
      int ret = slot->file->Read(buffer, nbytes);
      if (ret > 0) {
	if(!kernel->memoryManager->WriteBytes(buffer, 
					      ret, 
					      kernel->currentThread->space, 
					      (int)bp)) {
	  ret = -1;
	}
      }
      delete [] buffer;
      return (ret);
    }
  }
  return (-1);
}

#ifdef P3_3

/* method that returns the length of a file, basically
   a wrapper around OpenFile::Length()
 */

int FileIO::FileLength(OpenFileId fd) {
  if (fd < 0) return -1;

  struct OpenFileSlot *slot;

  for (int i=0; i < NFILES ; i++) {
    slot = &openFileSlot[i];
    if (slot->fd == fd) 
      return slot->file->Length();
  }

  return -1;
}

void FileIO::FileSeek(OpenFileId fd, int position) {
  if (fd < 0) return;

  struct OpenFileSlot *slot;

  for (int i=0; i < NFILES ; i++) {
    slot = &openFileSlot[i];
    if (slot->fd == fd) {
      slot->file->Seek(position);
      break;
    }
  }
}

#endif /* P3_3 */

#ifdef P4_1
OpenFileId FileIO::FileCreate(char *filename)
{

  // find an unused slot
  int i;
  for (i = 0; i < NFILES; i++) {
    if (openFileSlot[i].file == NULL)
      break;
  }
  
  if (i == NFILES) { // too many open files
    printf("Create %s: too many open files\n", filename);
    return (-1);
  }
  
  struct OpenFileSlot *slot = &openFileSlot[i];
  
  // Try to create the file
  
  slot->file = kernel->fileSystem->Create(currentDir, filename);
  if (slot->file == NULL) {
    printf("Create %s: cannot create\n", filename);
    return (-1);
  }
  slot->name = new char [strlen(filename) + 1];
  strcpy(slot->name, filename);
  printf("FileIO: Create %s: ID=%d\n", filename, i);
  return i;
}

int FileIO::FileRemove(char *filename)
{
    return (kernel->fileSystem->Remove(currentDir, filename));
}

int FileIO::FileStat(OpenFileId fd, filestat_t *stat)
{
  if (fd < 0 || fd >= NFILES || openFileSlot[(int)fd].file == NULL) {
        printf("Stat %d: invalid ID\n", fd);
        return (-1);
  }
  openFileSlot[(int)fd].file->Stat(stat);
  return (0);
}

int FileIO::ChDir(char *pathname)
{
    Directory *newDir = kernel->fileSystem->ChDir(currentDir, pathname);
    if (newDir) {
      //delete currentDir;
      currentDir = newDir;
      printf("ChDir %s: directory changed to #%d\n",
	     pathname, newDir->GetSector());
      strcpy(pwd,pathname);
      return (0);
    }
    printf ("ChDir %s, failed\n.",pathname);
    return (-1);
}

int FileIO::MkDir(char *pathname)
{
    bool result = kernel->fileSystem->MkDir(currentDir, pathname);
    if (result) return (0);
    return (-1);
}

#endif

