#include "thread.h"
#include "synch.h"
#include "list.h"
#include "hash.h"
#include "mmanager.h"

extern "C" void bcopy(const void *, void *, unsigned int);

/* Helper functions for the memory manager */

#ifdef P3_1
static int VpageKeyFunc(VirtPage *vp) { return vp->virtnum; }
#else
static int PpageKeyFunc(VirtPage *vp) { return vp->physnum; }
#endif /* P3_1 */
static unsigned int VpageHashFunc(int key) { return (unsigned int) key; }

//
// PageTable 
//

PageTable::PageTable(AddrSpace *_owner, int nPTEs) {
  owner = _owner;
#ifndef P3_1
  pageTable = new PTE [nPTEs];

  /* initalize the enteries in the page table to null */

  if (pageTable) {
    for (int i = 0; i < nPTEs; i++)
      pageTable[i] = -1;             /* vpn #i has ppn # = -1 */
  }

  PTEentries = nPTEs;
#else
  vp2ppTable = new HashTable<int,VirtPage *>(VpageKeyFunc,VpageHashFunc);
#endif /* ! P3_1 */
}

PageTable::~PageTable() {

#ifdef P3_1
#ifndef P3_2
  HashIterator<int, VirtPage *> *hi = 
    new HashIterator<int,VirtPage *>(vp2ppTable);
  List<int> *vlist = new List<int>;

  while (!hi->IsDone()) {
    vlist->Append(hi->Item()->virtnum);
    hi->Next();
  }

  while (!vlist->IsEmpty())
    vp2ppTable->Remove(vlist->RemoveFront());

  delete hi;
#endif /* ! P3_2 */
  delete vp2ppTable;
#else

 delete [] pageTable;

#endif /* P3_1 */

}


/* Set up a page table entry with logical-to-physical information. */

void PageTable::Enter(int virtPageNum, int physPageNum) {

#ifdef P3_1

  VirtPage *v = new VirtPage(physPageNum,virtPageNum,this);
  if(!vp2ppTable->IsInTable(virtPageNum))vp2ppTable->Insert(v);

#else

  ASSERT(virtPageNum < PTEentries);
  pageTable[virtPageNum] = physPageNum;  

#endif /* P3_1 */

}

/* Given the virtual page #, returns the physical page # of the page.
   It is up to the caller to verify first if the page is valid by
   calling IsValid(). */

int PageTable::PhysPageNum(int virtPageNum) {

#ifndef P3_1

  ASSERT(virtPageNum < PTEentries);
  return (pageTable[virtPageNum] & PN_MASK); 

#else

  VirtPage *v;
  int ppn;

  if(vp2ppTable->Find(virtPageNum,&v)) return v->physnum;
  if(kernel->memoryManager->iptm->LoadVPInMemory(owner, virtPageNum, &ppn)) return ppn;
  return NOSUCHPAGE;

#endif

}

/* return non-zero if the page is valid */

int PageTable::IsValid(int virtPageNum) {
#ifdef P3_1
#ifdef P3_2
  return (PhysPageNum(virtPageNum) >= 0);
#else
  return vp2ppTable->IsInTable(virtPageNum);
#endif
#else
  return (pageTable[virtPageNum] == -1 ? FALSE : TRUE);  
#endif /* P3_1 */
}

#ifndef P3_1
int PageTable::IsSet(int virtPageNum, int bit) {
  ASSERT(virtPageNum < PTEentries);
  return (pageTable[virtPageNum] & bit);
}
#endif /* ! P3_1 */

//
// Memory Manager
//

MemoryManager::MemoryManager(void)
{
  
#ifndef P3_1
  /* tell the machine that we will manage pageTable */

  kernel->machine->pageTable = new TranslationEntry[NumPhysPages];
#endif

  /* initalize the free page list */

  freePageList = new List<VirtPage *>;

  for (int i = 0; i < NumPhysPages; i++) {

#ifndef P3_1
    kernel->machine->pageTable[i].valid = FALSE;
#endif
    VirtPage *vPage = new VirtPage(i,-1,NULL);

    /* if the virtual page was created correctly, add it to the free
       page list, otherwise present an error */

    if (vPage == NULL) {
      printf("MemoryManager: Cannot put phys page #%x into free list\n", i);
    } else {
      freePageList->Append(vPage); 
    }

  }

  mutex = new Lock("MemoryManager");

#ifndef P3_1
  /* create the list of used pages */

  usedPageList = new HashTable<int,VirtPage *>(PpageKeyFunc, VpageHashFunc);
#else
#ifdef P3_1_DEBUG
  printf("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
  printf("Constructing IPTManager\n");
  printf("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
#endif
#ifdef P3_2
  iptm = new IPTManager(this);
#else
  iptm = new IPTManager;
#endif
  ASSERT(iptm->Constructed());

#endif /* ! P3_1 */

  printf("physical memory size: %d pages\n", NumPhysPages);
}

MemoryManager::~MemoryManager(void) {
  if(kernel->machine->pageTable) delete [] kernel->machine->pageTable;
  while (!freePageList->IsEmpty()) delete freePageList->RemoveFront();
  delete freePageList;
#ifndef P3_1
  delete usedPageList;
#else
  delete iptm;
#endif
  delete mutex;
}

// Allocate a page table on behalf on AddrSpace 'owner'.
// The page table will be large enough to contain 'VirtSize' bytes.

PageTable * MemoryManager::AllocatePageTable(AddrSpace *owner, int VirtPages)
{

  /* create a new page table */

#ifndef P3_1
  /**************************************
  ** VM implementation should allow programs larger than phy-mem
  **************************************/
  int numfree = freePageList->NumInList();
  if(numfree < VirtPages) {
    printf("MemoryManager::AllocatePageTable() - No free memory\n");
    return (PageTable*)NULL;
  }
#endif

  PageTable *table = new PageTable(owner, VirtPages);

#ifndef P3_1
  /* if the table was not correctly created, return NULL */

  if (table->Constructed() == FALSE) {
    printf("MemoryManager: cannot allocate page table\n");
    delete table;
    return (NULL);
  }

  /* The table is okay, return it */

  DEBUG('a',("MemoryManager: new page table, %d vpages\n", VirtPages));
#endif /* ! P3_1 */

  return (table);

}

#ifdef P3_3

/* free a particular page in the page table */

void MemoryManager::ReleasePage(PageTable *table, int vpn) {

  mutex->Acquire();

  HashTable<int,VirtPage *> *ht = table->vp2ppTable;
  int tmp_pid = table->owner->PID();
  VirtPage *v;
  
  if (ht->Find(vpn,&v)) {
    v = ht->Remove(vpn);
    iptm->Remove(IPTIndex(v->virtnum,tmp_pid));

    v->status = FREE;
    v->owner = NULL;
    v->virtnum = -1;

    freePageList->Append(v);    
  }

  mutex->Release();
}

#endif /* P3_3 */

// Free a page table, reclaiming all physical pages, if any.

void MemoryManager::ReleasePageTable(PageTable *table)
{

  // first acquire a lock, so all free'ing is atomic

  mutex->Acquire();

#ifdef P3_1

  HashTable<int,VirtPage *> *ht = table->vp2ppTable;
  HashIterator<int, VirtPage *> *hi = 
    new HashIterator<int,VirtPage *>(ht);

#ifdef P3_2
  
  List<int> *vlist = new List<int>;
  int tmp_pid = table->owner->PID();

  while (!hi->IsDone()) {
    vlist->Append(hi->Item()->virtnum);
    hi->Next();
  }
   
  while (!vlist->IsEmpty()) {
    VirtPage *v = ht->Remove(vlist->RemoveFront());
    iptm->Remove(IPTIndex(v->virtnum,tmp_pid));

    v->status = FREE;
    v->owner = NULL;
    v->virtnum = -1;

    freePageList->Append(v);
  }

  iptm->Remove(tmp_pid);

  delete vlist;

#else

  while (!hi->IsDone()) {
    VirtPage *v = hi->Item();
    iptm->Remove(IPTIndex(v->virtnum,v->physnum));

    v->status = FREE;
    v->owner = NULL;
    v->virtnum = -1;

    freePageList->Append(v);

    hi->Next();
  }

#endif /* P3_2 */

  delete hi;

#else

  for (int VpageNum = 0; VpageNum < table->GetLength(); VpageNum++) {

    // if page is not valid (i.e., not allocated), skip it.

    if (!table->IsValid(VpageNum)) {
      DEBUG('a',("mmanager: skipping vpage #%x\n", VpageNum));
      continue;
    }

    VirtPage *vPage;

    DEBUG('a',("MemoryManager: freeing vpage #%x (phys #%x)\n",
	  VpageNum,
	  table->PhysPageNum(VpageNum)));

    /* get the physical page for the virtual page we want to to free */

    int physPageNum = table->PhysPageNum(VpageNum);

    /* remove the page from the used list */

    vPage = (VirtPage *)usedPageList->Remove(physPageNum);

    /* if the page is used, reclaim it */

    ASSERT(vPage != NULL);
    vPage->status = FREE;                 /* the page is free for use */
    vPage->owner = NULL;                  /* the page is not owned by any
					     address space */
    kernel->machine->pageTable[VpageNum].valid = FALSE;
    vPage->virtnum = -1;                  
    freePageList->Append(vPage);   /* put this page on the free list */
  }

#endif /* P3_1 */

  /* done with our update, release the lock and delete the table */

  mutex->Release();
  delete table;
}

// Allocate a physical page on behalf of the page table owner,
// and map it onto its virtual address v.
// Return true if success.

bool MemoryManager::AllocatePhysPage(PageTable *table, int VirtPageNum)
{
#ifndef P3_1
  /* make sure all updates are atomic */

  mutex->Acquire();
  /* get the first free page */

  VirtPage *vPage = (VirtPage *)freePageList->RemoveFront();

  /* If the page that was returned is null we ran out of
     memory. Report the error and return */

  if (vPage == NULL) {
    printf("MemoryManager: ran out of memory\n");
    mutex->Release();
    return (FALSE);
  }
  vPage->virtnum = VirtPageNum;
  vPage->owner = table;
  vPage->status = BUSY;

  DEBUG('a',("MemoryManager: new vpage #%x (phys #%x)\n",
	VirtPageNum,
	vPage->physnum));

  usedPageList->Insert(vPage);
  table->Enter(VirtPageNum, vPage->physnum);
  kernel->machine->pageTable[VirtPageNum].valid = TRUE;
  kernel->machine->pageTable[VirtPageNum].virtualPage = VirtPageNum;
  kernel->machine->pageTable[VirtPageNum].physicalPage = vPage->physnum;

  mutex->Release();
  return (TRUE);
#else //P3_1
  /* make sure all updates are atomic */

  mutex->Acquire();
  /* get the first free page */

  if(!freePageList->IsEmpty()) {
    VirtPage *vPage = (VirtPage *)freePageList->RemoveFront();
    vPage->virtnum = VirtPageNum;
    vPage->owner = table;
    vPage->status = BUSY;
  
    DEBUG('a',("memoryManager: new vpage #%x (phys #%x)\n",
  	VirtPageNum,
  	vPage->physnum));
  
    TEWithPid te;
    te.pid = kernel->currentThread->pid;
    te.virtualPage = VirtPageNum;
    te.physicalPage = vPage->physnum;
    te.valid = TRUE;
    te.readOnly = FALSE;
    te.use = TRUE;
    iptm->Add(te);
  
    table->Enter(VirtPageNum, vPage->physnum);
  }
  else{
    int ppn;
    if(!iptm->FreePageInMemory(kernel->currentThread->space, VirtPageNum, &ppn)){
      printf("MemoryManager::AllocatePhysPage(pid=%d, vpn=%d) Failed to load page\n", 
	kernel->currentThread->pid, VirtPageNum);
      mutex->Release();
      return FALSE;
    }
    else{
      table->Enter(VirtPageNum, ppn);
    }
  }

  mutex->Release();
  return (TRUE);
#endif //P3_1
}

//----------------------------------------------------------------------
// MemoryManager::ReadMem
//      Read "size" (1, 2, or 4) bytes of virtual memory at "addr" into 
//	the location pointed to by "value".
//
//   	Returns FALSE if the translation step from virtual to physical memory
//   	failed.
//      "addrsp" -- the virt addr space for the user-prog
//	"addr" -- the virtual address to read from
//	"size" -- the number of bytes to read (1, 2, or 4)
//	"value" -- the place to write the result
//----------------------------------------------------------------------

bool
MemoryManager::ReadMem(AddrSpace* addrsp, int addr, int size, int *value)
{
    int data;
    bool exception;
    int physicalAddress;
    
    DEBUG(dbgAddr,("Reading VA 0x%x, size 0x%x\n", addr, size));
    
    exception = Translate(addrsp, addr, &physicalAddress, size, FALSE);
    if (!exception ) {
	return FALSE;
    }
    switch (size) {
      case 1:
	data = kernel->machine->mainMemory[physicalAddress];
	*value = data;
	break;
	
      case 2:
	data = *(unsigned short *) &kernel->machine->mainMemory[physicalAddress];
	*value = ShortToHost(data);
	break;
	
      case 4:
	data = *(unsigned int *) &kernel->machine->mainMemory[physicalAddress];
	*value = WordToHost(data);
	break;

      default: ASSERT(FALSE);
    }
    
    DEBUG(dbgAddr,("\tvalue read = 0x%x\n", *value));
    return (TRUE);
}

//----------------------------------------------------------------------
// MemoryManager::WriteMem
//      Write "size" (1, 2, or 4) bytes of the contents of "value" into
//	virtual memory at location "addr".
//
//   	Returns FALSE if the translation step from virtual to physical memory
//   	failed.
//
//      "addrsp" -- the virt addr space for the user-prog
//	"addr" -- the virtual address to write to
//	"size" -- the number of bytes to be written (1, 2, or 4)
//	"value" -- the data to be written
//----------------------------------------------------------------------

bool
MemoryManager::WriteMem(AddrSpace* addrsp, int addr, int size, int value)
{
    bool exception;
    int physicalAddress;
     
    DEBUG(dbgAddr,("Writing VA 0x%x, size 0x%x, value 0x%x\n", addr, size, value));

    exception = Translate(addrsp, addr, &physicalAddress, size, TRUE);
    if (!exception ) {
	return FALSE;
    }
    switch (size) {
      case 1:
	kernel->machine->mainMemory[physicalAddress] = (unsigned char) (value & 0xff);
	break;

      case 2:
	*(unsigned short *) &kernel->machine->mainMemory[physicalAddress]
		= ShortToMachine((unsigned short) (value & 0xffff));
	break;
      
      case 4:
	*(unsigned int *) &kernel->machine->mainMemory[physicalAddress]
		= WordToMachine((unsigned int) value);
	break;
	
      default: ASSERT(FALSE);
    }
    
    return TRUE;
}

//----------------------------------------------------------------------
// MemoryManager::Translate
// 	Translate a virtual address into a physical address, using 
//	either a page table or a TLB.  Check for alignment and all sorts 
//	of other errors, and if everything is ok, 
//	store the translated physical address in "physAddr".  
//      If there was an error, returns FALSE.
//
//      "addrsp" -- the virt addr space for the user-prog
//	"virtAddr" -- the virtual address to translate
//	"physAddr" -- the place to store the physical address
//	"size" -- the amount of memory being read or written
// 	"writing" -- if TRUE, check the "read-only" bit in the TLB
//----------------------------------------------------------------------

bool
MemoryManager::Translate(AddrSpace* addrsp, 
			 int virtAddr, 
			 int* physAddr, 
			 int size, 
			 bool writing)
{
    unsigned int vpn, offset;
    int ppn;

#ifndef P3_1
    VirtPage *vptr;
#endif

    DEBUG(dbgAddr,("\tTranslate 0x%x, %s\n", virtAddr, (writing ? "write" : "read")));

    if(!addrsp) return FALSE;

// check for alignment errors
    if (((size == 4) && (virtAddr & 0x3)) || ((size == 2) && (virtAddr & 0x1))){
	DEBUG(dbgAddr,("Alignment problem at 0x%x, size 0x%x", virtAddr, size));
	return FALSE;;
    }
    
    vpn = (unsigned) virtAddr / PageSize;
    offset = (unsigned) virtAddr % PageSize;
    PageTable* pt = addrsp->m_PageTable;
    if(!pt) {
      printf("Error!!! In MemoryManager::Translate() - Process Page Table missing!\n");
      return FALSE;
    }
#ifndef P3_1

    /* calculate the virtual page number, and offset within the page,
       from the virtual address */

    if(vpn >= pt->PTEentries) return FALSE;
    ppn = pt->pageTable[vpn];
    if(!usedPageList->Find(ppn, &vptr)) return FALSE;
    *physAddr = vptr->physnum * PageSize + offset;

    if( !((*physAddr >= 0) && ((*physAddr + size) <= MemorySize))) return FALSE;
    DEBUG(dbgAddr,("phys addr = 0x%x\n", *physAddr));
    return TRUE;
#else

    /* 
       Check if the page is in phys memory.  

       If the page is in phys memory, there will be a corresponding
       entry in process' m_PageTable.
    
       If not, get it from the disk.

       As pages are brought in & out of phys memory the m_PageTable of
       all the affected processes must be updated. For now, if the
       page is not in the process' m_PageTable, notify IPTManager, and
       let it do the rest. In case of bogus addresses, return FALSE.
    */


    if(!addrsp->VPageExistsInMainMemory(vpn, &ppn) ) {
      //page not found in main memory
      //notify  iptmanager
      //if iptmanager cannot locate the page, return FALSE
      /* printf("AddrSpace - \"%s\", VPN# %d Not found in Main Memory\n", 
	 (char*)addrsp->m_Name, vpn); */
      bool retval = iptm->LoadVPInMemory(addrsp, vpn, &ppn);
      if(!retval) {
        /* printf("AddrSpace - \"%s\", VPN# %d Not found even on disk\n", 
	   (char*)addrsp->m_Name, vpn); */
	return FALSE;
      }
      *physAddr = ppn * PageSize + offset;
      DEBUG(dbgAddr,("phys addr = 0x%x\n", *physAddr));
      if( !((*physAddr >= 0) && ((*physAddr + size) <= MemorySize))) return FALSE;
      return TRUE;
    } else {
      *physAddr = ppn * PageSize + offset;
      DEBUG(dbgAddr,("phys addr = 0x%x\n", *physAddr));
      if( !((*physAddr >= 0) && ((*physAddr + size) <= MemorySize))) return FALSE;
      return TRUE;
    }
	
#endif /* ! P3_1 */
}

/* MemoryManager::CopyStr
 *
 *    Copies a string from a virtual address (virtaddr) in a given
 *    address space (addrsp) into the given buffer (destbuf) 
 */

char * MemoryManager::CopyStr(char* destbuf, AddrSpace* addrsp, int virtaddr)
{
    char* data;
    bool exception;
    int physicalAddress;
    
    DEBUG(dbgAddr,("Reading VA 0x%x, size 0x%x\n", virtaddr, 1));
    
    exception = Translate(addrsp, virtaddr, &physicalAddress, 1, FALSE);
    if (!exception ) {
	return (char *)NULL;
    }
    data =  & kernel->machine->mainMemory[physicalAddress];
    strcpy(destbuf, data);
    
    DEBUG(dbgAddr,("\tstring read = %s\n", destbuf));
    return destbuf;
}

/*  MemoryManager::ReadBytes()
 *
 *     Assumes that buf has been allocated by the callee
 */

bool MemoryManager::ReadBytes(char* buf, int numBytes, AddrSpace* addrsp, int virtaddr)
{
    char* data;
    bool exception;
    int physicalAddress;
    int overflowAddress;
    
    DEBUG(dbgAddr,("Reading VA 0x%x, size 0x%x\n", virtaddr, 1));
    
    exception = Translate(addrsp, virtaddr, &physicalAddress, 1, FALSE);
    if (!exception ) {
	return FALSE;
    }

    exception = Translate(addrsp, numBytes + virtaddr, &overflowAddress, 1, FALSE);
    if (!exception ) {
	return FALSE;
    }

    data =  & kernel->machine->mainMemory[physicalAddress];
    bcopy(data, buf, numBytes);
    
    return TRUE;
}

/* MemoryManager::WriteBytes()
 *
 *    Copies buf to the virtaddr specified
 */

bool MemoryManager::WriteBytes(char* buf, 
			       int numBytes, 
			       AddrSpace* addrsp, 
			       int virtaddr)
{
    char* data;
    bool exception;
    int physicalAddress;
    int overflowAddress;
    
    DEBUG(dbgAddr,("Writing VA 0x%x, size 0x%x\n", virtaddr, 1));
    
    exception = Translate(addrsp, virtaddr, &physicalAddress, 1, FALSE);
    if (!exception ) {
	return FALSE;
    }

    exception = Translate(addrsp, numBytes + virtaddr, &overflowAddress, 1, FALSE);
    if (!exception ) {
	return FALSE;
    }

    data =  & kernel->machine->mainMemory[physicalAddress];
    bcopy(buf, data, numBytes);
    
    return TRUE;
}

#ifdef P3_1
//return SUCCESS(TRUE)/FAILURE(FALSE)
bool MemoryManager::HandleTlbMiss(int proc, int vpn, TEWithPid* tePtr) {

  //Procure the page from IPTManager
  //How IPTManager provides it (i.e. from memory/disc, is upto the IPTManager)
  //somehow, IPTManager will bring the page into the main memory
  //now find an entry from TLB to use 
  //if there is free entry, use it, else choose one to evict (for now FIFO)

  //printf("Entering MemoryManager::HandleTlbMiss(pid=%d, vpn=%d)\n", proc,vpn);
  //TEWithPid te;
  if(! iptm->HandleTlbMiss(proc, vpn, tePtr)) {
    printf("MemoryManager::HandleTlbMiss(pid=%d, vpn=%d) returning IPTManager FAILURE\n", proc, vpn);
    return FALSE;
  }
  //expect te.valid set to TRUE by this func

  int i = 0;
  for( i = 0; i < TLBSize; i++){
    if(!kernel->machine->tlb[i].valid) {
      break;
    }
  }

  if(i<TLBSize){

    /* found a free entry in TLB */

    kernel->machine->tlb[i].virtualPage = tePtr->virtualPage;
    kernel->machine->tlb[i].physicalPage= tePtr->physicalPage;

    /* Valid should be set to TRUE by iptm->HandleTlbMiss() */

    kernel->machine->tlb[i].valid       = tePtr->valid;     

    kernel->machine->tlb[i].readOnly    = tePtr->readOnly;
    kernel->machine->tlb[i].use         = tePtr->use;
    kernel->machine->tlb[i].dirty       = tePtr->dirty;

  } else{

    //choose an entry in TLB to evict 
    //currently using FIFO
    //0-th entry is the oldest
    //move all others 1 step up
    //use the last one for assignment
    //if this page is dirty, inform the IPTManager
    //IPTManager will write it to disk when it needs to flush it out of the memory

    iptm->MarkDirty(kernel->currentThread->pid, kernel->machine->tlb[0].virtualPage);
    for(int j = 0; j < TLBSize - 1; j++) {
      kernel->machine->tlb[j] = kernel->machine->tlb[j+1];
    }
    
    kernel->machine->tlb[TLBSize-1].virtualPage = tePtr->virtualPage;
    kernel->machine->tlb[TLBSize-1].physicalPage= tePtr->physicalPage;

    /* Valid should be set to TRUE by iptm->HandleTlbMiss() */

    kernel->machine->tlb[TLBSize-1].valid       = tePtr->valid; 

    kernel->machine->tlb[TLBSize-1].readOnly   = tePtr->readOnly;
    kernel->machine->tlb[TLBSize-1].use         = tePtr->use;
    kernel->machine->tlb[TLBSize-1].dirty       = tePtr->dirty;
  }

  return TRUE;
}

#endif /* P3_1 */

