#ifndef MMANAGER_H
#define MMANAGER_H

#include "hash.h"
#include "thread.h"
#include "synch.h"
#include "list.h"
#include "machine.h"

#ifdef P3_1
#include "iptmanager.h"
#define NOSUCHPAGE -1
#endif

class Lock;
class Condition;
class AddrSpace;
class PageTable;
typedef int PTE;

#ifdef P3_1
class IPTManager;
class TEWithPid;
#endif

enum _PageStatus { FREE, BUSY } ; // page status
typedef _PageStatus PageStatus;

/* Virtual Page structure */

class VirtPage {
 public:
  
  VirtPage(int p,int v, PageTable *o) {
    physnum = p;
    virtnum = v;
    owner = o;
    status = FREE;
  }

  ~VirtPage() {}

  int physnum;		                 // physical page #
  int virtnum;		                 // virtual page #, if in use
  PageTable *owner;	                 // page table using this page.
  enum PageStatus { FREE, BUSY } status; // page status
};

/*  Page Table Manager */

class PageTable {

 public:
  PageTable(AddrSpace *_owner, int nPTEs);
  ~PageTable();

  void Enter(int virtPageNum, int physPageNum);
  int PhysPageNum(int virtPageNum);
  int IsValid(int virtPageNum) ;

  bool Constructed() { return (pageTable != 0 ? TRUE : FALSE); }

#ifndef P3_1
  PTE *GetBaseAddr() { return (pageTable); };
#endif

  int GetLength() { return (PTEentries); };
  AddrSpace *GetOwner() { return (owner); };
  VirtPage* Remove(int i) {return vp2ppTable->Remove(i);}
  friend class MemoryManager;

private:

#ifndef P3_1
  int IsSet(int virtPageNum, int bit);
#endif

#ifdef P3_1
  HashTable<int,VirtPage *> *vp2ppTable;
#endif

  PTE *pageTable;                  /* this is an array indexed on virtual-page-number */
  AddrSpace *owner;	           /* address space owning this page table */
  int PTEentries;

  const int VALID_MASK = 127;      /* 64 pages of phys memory */
  const int PN_MASK    = 127;      /* 64 pages of virt memory */
};

//
// Memory Manager.
//

class MemoryManager {

 public:

  MemoryManager(void);
  ~MemoryManager(void);

  // Allocates a page table on behalf on AddrSpace 'owner'.
  // The page table will be large enough to contain 'VirtPages' pages.

  PageTable * AllocatePageTable(AddrSpace *owner, int VirtPages);

#ifdef P3_1
  /***********************************
  The program can be very large with VM, larger than physical mem available
  it's upto the callee to check how much allocation was possible 
  class PageTable also needs to be modified to remember this
  ---VVN
  ***********************************/
  IPTManager* iptm;                              /* To manage IPT */
#endif /* P3_1 */

#ifdef P3_3
  void ReleasePage(PageTable *table, int vpn);
#endif

  // Free a page table, reclaiming all physical pages, if any.

  void ReleasePageTable(PageTable *table);

  // Allocate a physical page on behalf of the page table owner, and
  // map it onto virtual page v.  Return true if success.

  bool AllocatePhysPage(PageTable *table, int VirtPageNum);
  bool ReadMem(AddrSpace* addrsp, int addr, int size, int *value);
  bool WriteMem(AddrSpace* addrsp, int addr, int size, int value);
  bool Translate(AddrSpace* addrsp, int virtAddr, int* physAddr, int size, bool writing);
  char * CopyStr(char* destbuf, AddrSpace* addrsp, int virtaddr);
  bool ReadBytes(char* buf, int numBytes, AddrSpace* addrsp, int virtaddr);
  bool WriteBytes(char* buf, int numBytes, AddrSpace* addrsp, int virtaddr);

#ifdef P3_1
  bool HandleTlbMiss(int proc, int vpn, TEWithPid* tePtr);
#endif

 private:

  Lock *mutex;		                        /* the memory manager
						 can be invoked by
						 more than one process
						 at a time, so we need
						 to synchronize its
						 job. */

  List<VirtPage *> *freePageList;	      /* list of free
						 pages. as we only
						 need to do
						 allocate/release of
						 free pages, this list
						 does not need to be
						 hashed. */

#ifdef P3_1

  /* 

    Machine uses either PageTable or TLB for Address Translation.
    
    PageTable is essentially a mapping of VPN to corresponding PPN.
    The size of the pagetable is equal to the number of virtpages
    required for the process So the max size equals total number of
    phys pages in memory.

    In case of PageTable, Machine is only aware of the PageTable of
    the currently running process. On Context Switch, machine's
    pagetable points to the new process' pagetable.  Every process is
    aware of it's own PageTable.  This means that every process knows
    the phys pages used by itself and the machine knows only about the
    phys pages used by one process.
    
    MemoryManager must keep a tab on which pages are currently in use
    by all the processes.  This list is indexed on Phys Page Number to
    detect diff processes trying to access the same phys page. This
    list does not change till a new process is started, or a running
    process terminates.

    This list is inappropriate when VM is used as the physical pages
    used by a process keeps changing as pages are swapped in & out.
    Instead we need to manage an Inverse Page Table which gives a list
    of used phys pages indexed on a combination of pid & virtpage num.

    Page Replacement Algorithm will traverse through the hashtable
    using an iterator to choose the page to be evicted.  This logic is
    kept in IPTManager class so that MemoryManger does not have worry
    about it.

  */

  friend class IPTManager;

#else

  HashTable<int,VirtPage *> *usedPageList;       /* list of used
						    pages. we use a
						    hashed list, so
						    that we can easy
						    lookup the page we
						    are interested in.
						    PhysPageNum is
						    used as key */
#endif /* P3_1 */

};

#endif /* ! MMANAGER_H */


