#include "pidmanager.h"

/* Process Table helper functions */

static int PidGetFunc(PidInfo *info) { return info->pid; }
static unsigned int PidHashFunc(PID p) { return (unsigned int) p; }

PidManager::PidManager(void) {

  mutex = new Lock("PidManager");
  condition = new Condition("PidManager");
  procTable = new HashTable<PID,PidInfo *>(PidGetFunc,PidHashFunc);

  NextPid = 1;
  ActivePidCount = 0;

}

PidManager::~PidManager(void) {
  condition->Broadcast(mutex);       /* wake anyone who is waiting */
  delete mutex;
  delete condition;
  delete procTable;
}

/* PidManager::PidAllocate
 *
 *   Allocates a unique PID for this process. The PID is a
 *   monotonically increasing number until MAXPID is reached, at which
 *   point it wraps around at 0.  
 */

#ifdef P3_1
PID PidManager::PidAllocate(Thread *t) {
#else
PID PidManager::PidAllocate() {
#endif

  PidInfo *info;
  PID ret;
  mutex->Acquire();

  if (ActivePidCount >= MAX_ACTIVE_PIDS) {
    mutex->Release();
    return (NOPID);
  }

  while (procTable->IsInTable(NextPid)) {
    if (++NextPid > MAXPID) NextPid = 1;
  }

  info = new PidInfo;
  info->pid = NextPid;
  info->state = IN_USE;

#ifdef P3_1
  info->thread = t;
#endif

  procTable->Insert(info);
  ret = NextPid;
  if (++NextPid > MAXPID) NextPid = 1;
  
  ActivePidCount++;
  mutex->Release();
  return (ret);

}

/* PidManager::PidTerminate
 *
 *     Marks the given PID's state as ZOMBIE. Any processes WAITing on
 *     this PID are awakened.  
 */

void PidManager::PidTerminate(PID pid) {
  PidInfo *info;

  mutex->Acquire();  


  /* if the given PID is in the table, set it state to ZOMBIE, reduce
     the count of active processes and broadcast to everyone waiting
     on the condition variable */

  if (procTable->Find(pid,&info)) {
    ASSERT(info != NULL);
    info->state = ZOMBIE;

#ifdef P3_1
    procTable->Remove(pid);          /* remove the pid from the process table */
#endif

    ActivePidCount--;
    condition->Broadcast(mutex);
  } 

  mutex->Release();
}


/* PidManager::PidWait 
 *
 *    Wait for a process 'pid' to terminate. if the process does not
 *    exist, return error, otherwise wait until PID becomes zombie.  
 */

int PidManager::PidWait(PID pid) {

  if (pid <= NOPID || pid > MAXPID) {
    printf("PidManager: illegal PID #%d\n", pid);
    return (NOPID);
  }

  PidInfo *info;
  mutex->Acquire();

  while (TRUE) {

    if (!procTable->Find(pid,&info)) {
      //printf("Wait: PID #%d unexistent\n", pid);
      mutex->Release();
      return (NOPID);
    }

    if (info->state == IN_USE) {
      //printf("DEBUG: Waiting on pid %d.\n", pid);
      condition->Wait(mutex);
    } else {
      procTable->Remove(info->pid);
      delete info;
      mutex->Release();
      //printf("DEBUG: Done waiting on pid %d.\n", pid);
      return (pid);
    }
  }

  /* Not Reached */
}

/* PidManager::ActivePids
 * 
 *    Returns the number of active PIDs (running processes)
 */

int PidManager::ActivePids() { return (ActivePidCount); }

#ifdef P3_1
/* PidManager::getThread
 * 
 *    Returns a pointer to the thread corresponding to a given PID
 */

Thread * PidManager::getThread(PID pid) {

  if (pid <= NOPID || pid > MAXPID) {
    printf("getThread: illegal PID #%d\n", pid);
    return NULL;
  }

  PidInfo *info;
  mutex->Acquire();

  if (!procTable->Find(pid,&info)) {
    printf("getThread: PID #%d unexistent\n", pid);
    mutex->Release();
    return NULL;
  }

  mutex->Release();
  return info->thread;
}
#endif
